

import numpy as np
import scipy.constants as scpct
import datastock as ds


# ##############################################################
# ##############################################################
#              Maxwellian
# ##############################################################


def get_maxwellian(
    kTe_eV=None,
    mass=None,
    # output
    energy_eV=None,
    velocity_ms=None,
):
    """ Return a Maxwell-Boltzmann distribution

    Calculated to be:
        - analytically normalized
        - against the variable provided

    See:
        https://en.wikipedia.org/wiki/Maxwell%E2%80%93Boltzmann_distribution

    """

    # --------------
    # check input
    # --------------

    kTe_eV, mass, val, key = _check_maxwell(
        kTe_eV=kTe_eV,
        mass=mass,
        # output
        energy_eV=energy_eV,
        velocity_ms=velocity_ms,
    )

    # --------------
    # Compute
    # --------------

    if key == 'velocity_ms':

        dist = (
            (2./np.pi)
            * (mass / kTe_eV)**1.5
            * velocity_ms**2
            * np.exp(-mass*val**2 / (2 * kTe_eV))
        )

    elif key == 'energy_eV':

        dist = (
            2 * np.sqrt(energy_eV / np.pi)
            * (1./kTe_eV)**1.5
            * np.exp(-energy_eV / kTe_eV)
        )

    else:
        raise NotImplementedError()

    # --------------
    # format output
    # --------------

    dout = {
        'maxwell': {
            'vs': key,
            'data': dist,
            'units': None,
        },
    }

    return dout


def _check_maxwell(
    kTe_eV=None,
    mass=None,
    # output
    energy_eV=None,
    velocity_ms=None,
):

    # ------------------
    # energy vs velocity
    # ------------------

    dunique = {
        'energy_eV': energy_eV,
        'velocity_ms': velocity_ms,
    }

    lok = [k0 for k0, v0 in dunique.items() if v0 is not None]
    if len(lok) != 1:
        lstr = [f"\t- {k0}: v0" for k0, v0 in dunique.items()]
        msg = (
            "Please provide only one of the following:\n"
            + "\n".join(lstr)
        )
        raise Exception(msg)

    key = lok[0]
    val = dunique[key]

    # ------------
    # mass
    # ------------

    if key == 'velocity_ms':
        if mass is None:
            mass = scpct.m_e
    else:
        mass = None

    # ---------------
    # broadcastable
    # ---------------

    dout = {
        'kTe_eV': kTe_eV,
        key: val,
    }
    if key == 'velocity_ms':
        dout['mass'] = mass

    dout, shape = ds._generic_check._check_all_broadcastable(
        **dout,
    )

    lout = ['kTe_eV', key, 'mass']
    lout = [dout[k0] for k0 in lout if k0 in dout.keys()]
    if key == 'energy_eV':
        lout.append(None)
    lout.append(key)

    return lout



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import datastock as ds


# specific
from . import _generic_check
from . import _generic_plot


# ################################################################
# ################################################################
#                           plot check
# ################################################################


def _plot_diagnostic_check(
    coll=None,
    key=None,
    key_cam=None,
    # parameters
    vmin=None,
    vmax=None,
    alpha=None,
    dx0=None,
    dx1=None,
    # figure
    plot_colorbar=None,
    proj=None,
    data=None,
    units=None,
    rocking_curve=None,
    los_res=None,
    # interactivity
    color_dict=None,
    nlos=None,
    connect=None,
):

    # -------
    # key

    # key
    key, key_cam = coll.get_diagnostic_cam(key, key_cam)
    is2d = coll.dobj['diagnostic'][key]['is2d']
    # spectro = coll.dobj['diagnostic'][key]['spectro']

    # if spectro:
    #     assert len(key_cam) == 1
    #     doptics = coll.dobj['diagnostic'][key]['doptics'][key_cam[0]]
    #     kcryst = doptics['optics'][doptics['ispectro'][0]]
    # else:
    #    pass
    #    kcryst = None

    # -------
    # data

    # default
    defdata = 'etendue'
    c0 = (
        data is None
        and all([
            v0.get(defdata) is not None
            for v0 in coll.dobj['diagnostic'][key]['doptics'].values()
        ])
    )
    if c0:
        data = defdata

    # get data dict
    out = coll.get_diagnostic_data(
        key=key,
        key_cam=key_cam,
        data=data,
        units=units,
    )

    # extract
    if out is None:
        ddata, dref, units, static, daxis = {}, {}, None, True, None
    else:
        ddata, dref, units, static, daxis = out

    refz = None
    if static is False:
        lnr = [len(v0) for v0 in dref.values()]
        refz = [v0[daxis[k0]] for k0, v0 in dref.items()]

        if len(set(lnr)) != 1:
            msg = f"data '{data}' shall have the same ndims for all cameras!"
            raise Exception(msg)

        if len(set(refz)) != 1:
            msg = f"data '{data}' shall have the same extra ref for all cameras"
            raise Exception(msg)

        refz = refz[0]

    ylab = f"{data} ({units})"

    # ----------
    # vmin, vmax

    if vmin is None and len(ddata) > 0:
        vmin = np.nanmin([np.nanmin(v0) for v0 in ddata.values()])

    if vmax is None and len(ddata) > 0:
        vmax = np.nanmax([np.nanmax(v0) for v0 in ddata.values()])

    # -----
    # alpha

    alpha = ds._generic_check._check_var(
        alpha, 'alpha',
        types=float,
        default=0.2,
        sign='> 0.',
    )

    # -----
    # proj

    pall = ['cross', 'hor', '3d', 'camera', 'traces']
    proj = _generic_plot._proj(
        proj=proj,
        pall=pall,
    )

    if static is True:
        proj = [pp for pp in proj if pp != 'traces']

    # ----------
    # los_res

    los_res = ds._generic_check._check_var(
        los_res, 'los_res',
        types=float,
        default=0.05,
        sign='> 0.',
    )

    # -------
    # color_dict

    color_dict = _check_color_dict(color_dict)

    # -------
    # nlos

    nlos = ds._generic_check._check_var(
        nlos, 'nlos',
        types=int,
        default=5,
    )

    # ---------------
    # dx0, dx1

    # dx0
    dx0 = float(ds._generic_check._check_var(
        dx0, 'dx0',
        types=(int, float),
        default=0.,
    ))

    # dx1
    dx1 = float(ds._generic_check._check_var(
        dx1, 'dx1',
        types=(int, float),
        default=0.,
    ))

    # -------
    # plot_colorbar

    plot_colorbar = ds._generic_check._check_var(
        plot_colorbar, 'plot_colorbar',
        types=bool,
        default=True,
    )

    # -------
    # connect

    connect = ds._generic_check._check_var(
        connect, 'connect',
        types=bool,
        default=True,
    )

    return (
        key,
        key_cam,
        is2d,
        proj,
        ddata,
        dref,
        static,
        daxis,
        refz,
        vmin,
        vmax,
        alpha,
        units,
        los_res,
        color_dict,
        nlos,
        dx0,
        dx1,
        ylab,
        plot_colorbar,
        connect,
    )


def _check_color_dict(color_dict=None):
    if color_dict is None:
        lc = ['r', 'g', 'b', 'm', 'c', 'y']
        color_dict = {
            'x': lc,
            'y': lc,
        }
    return color_dict


# ##################################################################
# ##################################################################
#                           plot main
# ##################################################################


def _plot_diagnostic(
    coll=None,
    key=None,
    key_cam=None,
    keyZ=None,
    optics=None,
    elements=None,
    proj=None,
    los_res=None,
    # data plot
    data=None,
    units=None,
    cmap=None,
    vmin=None,
    vmax=None,
    alpha=None,
    dx0=None,
    dx1=None,
    # config
    plot_config=None,
    plot_colorbar=None,
    # figure
    dax=None,
    dmargin=None,
    fs=None,
    wintit=None,
    # interactivity
    color_dict=None,
    nlos=None,
    dinc=None,
    connect=None,
):

    # ------------
    # check inputs

    (
        key,
        key_cam,
        is2d,
        proj,
        ddata,
        dref,
        static,
        daxis,
        refz,
        vmin,
        vmax,
        alpha,
        units,
        los_res,
        color_dict,
        nlos,
        dx0,
        dx1,
        ylab,
        plot_colorbar,
        connect,
    ) = _plot_diagnostic_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # parameters
        vmin=vmin,
        vmax=vmax,
        alpha=alpha,
        dx0=dx0,
        dx1=dx1,
        # figure
        plot_colorbar=plot_colorbar,
        proj=proj,
        data=data,
        units=units,
        los_res=los_res,
        # interactivity
        color_dict=color_dict,
        nlos=nlos,
        connect=connect,
    )

    # ------------
    # prepare data

    dplot = coll.get_diagnostic_dplot(
        key=key,
        key_cam=key_cam,
        optics=optics,
        elements=elements,
        dx0=dx0,
        dx1=dx1,
    )

    # -------------------------
    # prepare los interactivity

    # instanciate new Datastock
    coll2 = coll.__class__()

    # ---------------------
    # prepare los and ddata

    # dcamref
    dcamref, drefx, drefy = _prepare_dcamref(
        coll=coll,
        key_cam=key_cam,
        is2d=is2d,
    )

    # los
    dlos_n, dref_los = _prepare_los(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        key_diag=key,
        key_cam=key_cam,
        los_res=los_res,
    )

    # vos
    dvos_n, dref_vos = _prepare_vos(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        key_diag=key,
        key_cam=key_cam,
        los_res=los_res,
    )

    # ddatax, ddatay
    _, dkeyx, dkeyy, ddatax, ddatay, dextent = _prepare_datarefxy(
        coll=coll,
        coll2=coll2,
        dcamref=dcamref,
        drefx=drefx,
        drefy=drefy,
        ddata=ddata,
        static=static,
        dx0=dx0,
        dx1=dx1,
        is2d=is2d,
    )

    # ---------------------
    # prepare non-static

    if static is False and len(ddata) > 0:

        k0 = key_cam[0]
        keyz = coll.get_ref_vector(ref=refz, key0=keyZ)[3]
        nz = ddata[k0].shape[daxis[k0]]

        keyz, zstr, dz2, labz = ds._generic_utils_plot._get_str_datadlab(
            keyX=keyz, nx=nz, islogX=False, coll=coll,
        )
        dataz = coll.ddata[keyz]['data']

        npts = 0
        for k0 in key_cam:
            npts = max(npts, ddata[k0].size)

        bck = 'envelop' if npts > 10000 else 'lines'

        coll2.add_ref(key=refz, size=nz)
        coll2.add_data(key=keyz, data=dataz, ref=refz)

        # add camera data
        for k0 in key_cam:
            coll2.add_data(
                key=f'{k0}_{data}',
                data=ddata[k0].T,
                ref=dref[k0][::-1],
                units=units,
            )

    # -----------------
    # prepare figure

    if dax is None:

        dax = _generic_plot.get_dax_diag(
            proj=proj,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            tit=key,
            is2d=is2d,
            key_cam=key_cam,
        )

    dax = _generic_check._check_dax(dax=dax, main=proj[0])

    # -----------------
    # plot static parts

    _plot_diag_geom(
        dax=dax,
        key_cam=key_cam,
        dplot=dplot,
        is2d=is2d,
    )

    # plot data
    if static is True:

        for k0 in key_cam:
            kax = f'{k0}_sig'
            if dax.get(kax) is not None:
                if ddata is None or ddata.get(k0) is None:
                    continue

                ax = dax[kax]['handle']

                if is2d:
                    im = ax.imshow(
                        ddata[k0].T,
                        extent=dextent[k0],
                        cmap=cmap,
                        vmin=vmin,
                        vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                    )
                    if plot_colorbar is True:
                        plt.colorbar(im, ax=ax)

                else:
                    ax.plot(
                        ddata[k0],
                        c='k',
                        ls='-',
                        lw=1.,
                        marker='.',
                        ms=6,
                    )
                    ax.set_xlim(-1, ddata[k0].size)
                    ax.set_ylabel(ylab)
                    ax.set_title(k0, size=12, fontweight='bold')

                    if vmin is not None:
                        ax.set_ylim(bottom=vmin)
                    if vmax is not None:
                        ax.set_ylim(top=vmax)

    else:
        # plot traces envelop
        for k0 in key_cam:

            kax = f'{k0}_trace'
            if dax.get(kax) is None or ddata.get(k0) is None:
                continue

            ax = dax[kax]['handle']

            if bck == 'lines':
                shap = list(ddata[k0].shape)
                shap[daxis[k0]] = 1
                bckl = np.concatenate(
                    (ddata[k0], np.full(shap, np.nan)),
                    axis=daxis[k0],
                )
                bckl = np.swapaxes(bckl, daxis[k0], -1).ravel()

                ax.plot(
                    np.tile(np.r_[dataz, np.nan], int(np.prod(shap))),
                    bckl,
                    c=(0.8, 0.8, 0.8),
                    ls='-',
                    lw=1.,
                    marker='None',
                )

            else:
                tax = tuple([
                    ii for ii in range(ddata[k0].ndim) if ii != daxis[k0]
                ])

                ax.fill_between(
                    dataz,
                    np.nanmin(ddata[k0], axis=tax),
                    np.nanmax(ddata[k0], axis=tax),
                    facecolor=(0.8, 0.8, 0.8, 0.8),
                    edgecolor='None',
                )

            if vmin is not None:
                ax.set_ylim(bottom=vmin)
            if vmax is not None:
                ax.set_ylim(top=vmax)

    # ----------------
    # define and set dgroup

    if coll2 is not None:
        dgroup = {
            f'{k0}_x': {
                'ref': [drefx[k0]],
                'data': ['index'],
                'nmax': nlos,
            }
            for k0 in key_cam
        }

        if is2d:
            dgroup.update({
                'y': {
                    'ref': list(drefy.values()),
                    'data': ['index'],
                    'nmax': nlos,
                },
            })

        if static is False is not None:
            dgroup.update({
                't': {
                    'ref': [refz],
                    'data': ['index'],
                    'nmax': 1,
                },
            })

    # ------------------
    # plot mobile parts

    for k0 in key_cam:

        if dlos_n[k0] is not None:
            nan_los = np.full((dlos_n[k0],), np.nan)
        else:
            nan_los = None

        if dvos_n is None:
            nan_vos = None
        else:
            nan_vos = np.full((dvos_n[k0],), np.nan)

        # cross
        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            _add_camera_los_cross(
                coll2=coll2,
                k0=k0,
                ax=ax,
                kax=kax,
                nlos=nlos,
                dref_los=dref_los,
                dref_vos=dref_vos,
                color_dict=color_dict,
                nan_los=nan_los,
                nan_vos=nan_vos,
                alpha=alpha,
            )

        # hor
        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            _add_camera_los_hor(
                coll2=coll2,
                k0=k0,
                ax=ax,
                kax=kax,
                nlos=nlos,
                dref_los=dref_los,
                dref_vos=dref_vos,
                color_dict=color_dict,
                nan_los=nan_los,
                nan_vos=nan_vos,
                alpha=alpha,
            )

        # 3d
        kax = '3d'
        if dax.get(kax) is not None and nan_los is not None:
            ax = dax[kax]['handle']

            for ii in range(nlos):
                l0, = ax.plot(
                    nan_los,
                    nan_los,
                    nan_los,
                    c=color_dict['x'][ii%len(color_dict['x'])],
                    ls='-',
                    lw=1.,
                )

                # add mobile
                kl0 = f'{k0}_los-3d-{ii}'
                # coll2.add_mobile(
                # key=kl0,
                # handle=l0,
                # refs=reflos,
                # data=['index', 'index', 'index'],
                # dtype=['xdata', 'ydata', 'zdata'],
                # axes=kax,
                # ind=ii,
                # )

        # camera
        kax = f'{k0}_sig'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']

            _add_camera_vlines_marker(
                coll2=coll2,
                dax=dax,
                ax=ax,
                kax=kax,
                is2d=is2d,
                k0=k0,
                nlos=nlos,
                ddatax=ddatax,
                ddatay=ddatay,
                drefx=drefx,
                drefy=drefy,
                dkeyx=dkeyx,
                dkeyy=dkeyy,
                color_dict=color_dict,
            )

        # vline on traces
        if static is False:

            kax = f'{k0}_trace'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                lv = ax.axvline(
                    dataz[0],
                    c='k',
                    lw=1.,
                    ls='-',
                )

                kv = f'{k0}_zline'
                coll2.add_mobile(
                    key=kv,
                    handle=lv,
                    refs=(refz,),
                    data=[keyz],
                    dtype=['xdata'],
                    axes=kax,
                    ind=0,
                )

                dax[kax].update(refx=[refz], datax=[keyz])

    # -------------------
    # data if not static

    if static is False:

        for k0 in key_cam:

            # line/im plot on data
            kax = f'{k0}_sig'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                tax = tuple([
                    ii for ii in range(ddata[k0].ndim) if ii != daxis[k0]
                ])

                if is2d:
                    im = ax.imshow(
                        np.take(ddata[k0], 0, axis=daxis[k0]).T,
                        extent=dextent[k0],
                        cmap=cmap,
                        vmin=vmin,
                        vmax=vmax,
                        origin='lower',
                        interpolation='nearest',
                    )
                    if plot_colorbar is True:
                        plt.colorbar(im, ax=ax)

                    km = f'{k0}_{data}'
                    coll2.add_mobile(
                        key=km,
                        handle=im,
                        refs=((refz,),),
                        data=[f'{k0}_{data}'],
                        dtype='data',
                        axes=kax,
                        ind=0,
                    )

                else:

                    l0, = ax.plot(
                        np.take(ddata[k0], 0, axis=daxis[k0]),
                        c='k',
                        ls='-',
                        lw=1.,
                        marker='.',
                        ms=6,
                    )
                    ax.set_xlim(-1, ddata[k0].size / nz)
                    ax.set_ylabel(ylab)
                    ax.set_title(k0, size=12, fontweight='bold')

                    if vmin is not None:
                        ax.set_ylim(bottom=vmin)
                    if vmax is not None:
                        ax.set_ylim(top=vmax)

                    km = f'{k0}_{data}'
                    coll2.add_mobile(
                        key=km,
                        handle=l0,
                        refs=((refz,),),
                        data=[f'{k0}_{data}'],
                        dtype='ydata',
                        axes=kax,
                        ind=0,
                    )

                    if vmin is not None:
                        ax.set_ylim(bottom=vmin)
                    if vmax is not None:
                        ax.set_ylim(top=vmax)

            # line plot on traces
            kax = f'{k0}_trace'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                sli = tuple([
                    slice(None) if ii == daxis[k0] else 0
                    for ii in range(ddata[k0].ndim)
                ])

                for ii in range(nlos):
                    l0, = ax.plot(
                        dataz,
                        ddata[k0][sli],
                        c=color_dict['x'][ii%len(color_dict['x'])],
                        lw=1.,
                        ls='-',
                    )

                    refi = dref_los[k0][0] if is2d else dref_los[k0][0]
                    kv = f'{k0}_trace{ii}'
                    coll2.add_mobile(
                        key=kv,
                        handle=l0,
                        refs=(refi,),
                        data=[f'{k0}_{data}'],
                        dtype=['ydata'],
                        axes=kax,
                        ind=ii,
                    )

    # -------
    # config

    if plot_config.__class__.__name__ == 'Config':

        kax = 'cross'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

        kax = 'hor'
        if dax.get(kax) is not None:
            ax = dax[kax]['handle']
            plot_config.plot(lax=ax, proj=kax, dLeg=False)

    # -------
    # connect

    if coll2.dobj.get('mobile') is not None:
        # add axes
        for ii, kax in enumerate(dax.keys()):
            harmonize = ii == len(dax.keys()) - 1
            coll2.add_axes(key=kax, harmonize=harmonize, **dax[kax])

        # connect
        if connect is True:
            coll2.setup_interactivity(
                kinter='inter0',
                dgroup=dgroup,
                dinc=dinc,
            )
            coll2.disconnect_old()
            coll2.connect()

            coll2.show_commands()
            return coll2
        else:
            return coll2, dgroup
    else:
        return dax


# ################################################################
# ################################################################
#                       Prepare
# ################################################################


def _prepare_dcamref(
    coll=None,
    key_cam=None,
    is2d=None,
):
    dcamref = {
        k0: coll.dobj['camera'][k0]['dgeom']['ref']
        for k0 in key_cam
    }
    drefx = {k0: v0[0] for k0, v0 in dcamref.items()}

    if is2d:
        drefy = {k0: v0[1] for k0, v0 in dcamref.items()}
    else:
        drefy = None

    return dcamref, drefx, drefy


def _prepare_los(
    coll=None,
    coll2=None,
    dcamref=None,
    key_diag=None,
    key_cam=None,
    los_res=None,
):

    # -----------------
    # create dlos, dvos

    # dlos
    dlos_n = {
        k0: coll.dobj['diagnostic'][key_diag]['doptics'][k0]['los']
        for k0 in key_cam
    }
    dref_los = {}

    # -------------
    # los

    # los on cams
    for k0, v0 in dcamref.items():
        for rr in v0:
            coll2.add_ref(key=rr, size=coll.dref[rr]['size'])

        # los
        if dlos_n[k0] is not None:

            los_x, los_y, los_z = coll.sample_rays(
                key=dlos_n[k0],
                res=los_res,
                mode='rel',
                concatenate=False,
            )
            los_r = np.hypot(los_x, los_y)
            reflos = coll.dobj['rays'][dlos_n[k0]]['ref']
            dref_los[k0] = (reflos[1:], reflos[1:])

            if reflos[0] not in coll2.dref.keys():
                coll2.add_ref(key=reflos[0], size=los_x.shape[0])

            coll2.add_data(key=f'{k0}_los_x', data=los_x, ref=reflos)
            coll2.add_data(key=f'{k0}_los_y', data=los_y, ref=reflos)
            coll2.add_data(key=f'{k0}_los_z', data=los_z, ref=reflos)
            coll2.add_data(key=f'{k0}_los_r', data=los_r, ref=reflos)

            # store x, y, z
            dlos_n[k0] = los_x.shape[0]

    return dlos_n, dref_los


def _prepare_vos(
    coll=None,
    coll2=None,
    dcamref=None,
    key_diag=None,
    key_cam=None,
    los_res=None,
):

    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    if doptics[key_cam[0]].get('dvos') is None:
        return None, None

    # -----------------
    # create dlos, dvos

    # dvos
    dvos_n = {
        k0: {'pc': doptics[k0]['dvos']['pcross']}
        for k0 in key_cam
    }
    dref_vos = {}

    # ----
    # vos

    # vos on cams
    for k0, v0 in dcamref.items():

        krxy = f'{k0}_xy2'
        coll2.add_ref(key=krxy, size=2)

        # vos
        if dvos_n[k0]['pc'] is not None:

            pc0 = coll.ddata[dvos_n[k0]['pc'][0]]['data']
            pc1 = coll.ddata[dvos_n[k0]['pc'][1]]['data']
            pcref = coll.ddata[dvos_n[k0]['pc'][0]]['ref']

            ph0, ph1, phref = None, None, None
            if doptics[k0]['dvos'].get('phor') is not None:
                ph0 = coll.ddata[doptics[k0]['dvos']['phor'][0]]['data']
                ph1 = coll.ddata[doptics[k0]['dvos']['phor'][1]]['data']
                phref = coll.ddata[doptics[k0]['dvos']['phor'][0]]['ref']

            if pcref[0] not in coll2.dref.keys():
                coll2.add_ref(key=pcref[0], size=pc0.shape[0])

            dref_vos[k0] = (pcref[1:],)

            ref = tuple(list(pcref[::-1]) + [krxy])
            pcxy = np.array([pc0, pc1]).T
            coll2.add_data(key=f'{k0}_vos_cross', data=pcxy, ref=ref)
            if ph0 is not None:
                ref = tuple(list(phref[::-1]) + [krxy])
                phxy = np.array([ph0, ph1]).T
                coll2.add_data(key=f'{k0}_vos_hor', data=phxy, ref=ref)

            # store
            dvos_n[k0] = pc0.shape[0]

    return dvos_n, dref_vos


def _prepare_datarefxy(
    coll=None,
    coll2=None,
    dcamref=None,
    drefx=None,
    drefy=None,
    ddata=None,
    static=None,
    dx0=None,
    dx1=None,
    is2d=None,
):

    # ---------------
    # dx0, dx1

    # dx0
    dx0 = float(ds._generic_check._check_var(
        dx0, 'dx0',
        types=(int, float),
        default=0.,
    ))

    # dx1
    dx1 = float(ds._generic_check._check_var(
        dx1, 'dx1',
        types=(int, float),
        default=0.,
    ))

    # -----------------
    # prepare

    # prepare dict
    dkeyx, ddatax = {}, {}
    if is2d:
        dkeyy, ddatay, dextent = {}, {}, {}
    else:
        dkeyy, ddatay, dextent = None, None, None

    # loop on cams
    for k0, v0 in dcamref.items():

        # datax, datay
        if ddata is not None:
            if is2d:
                dkeyx[k0], dkeyy[k0] = coll.dobj['camera'][k0]['dgeom']['cents']

                ddatax[k0] = coll.ddata[dkeyx[k0]]['data'] + dx0
                ddatay[k0] = coll.ddata[dkeyy[k0]]['data'] + dx1

                coll2.add_data(key=dkeyx[k0], data=ddatax[k0], ref=drefx[k0])
                coll2.add_data(key=dkeyy[k0], data=ddatay[k0], ref=drefy[k0])
            else:
                dkeyx[k0] = f'{k0}_i0'
                ddatax[k0] = np.arange(0, coll.dref[drefx[k0]]['size']) + dx0
                coll2.add_data(key=dkeyx[k0], data=ddatax[k0], ref=drefx[k0])

            # -------------------------
            # extent

            reft = None
            if is2d:
                if ddatax[k0].size == 1:
                    ddx = coll.ddata[coll.dobj['camera'][k0]['dgeom']['outline'][0]]['data']
                    ddx = np.max(ddx) - np.min(ddx)
                else:
                    ddx = ddatax[k0][1] - ddatax[k0][0]
                if ddatay[k0].size == 1:
                    ddy = coll.ddata[coll.dobj['camera'][k0]['dgeom']['outline'][1]]['data']
                    ddy = np.max(ddy) - np.min(ddy)
                else:
                    ddy = ddatay[k0][1] - ddatay[k0][0]

                dextent[k0] = (
                    ddatax[k0][0] - 0.5*ddx,
                    ddatax[k0][-1] + 0.5*ddx,
                    ddatay[k0][0] - 0.5*ddy,
                    ddatay[k0][-1] + 0.5*ddy,
                )

    return reft, dkeyx, dkeyy, ddatax, ddatay, dextent


# ################################################################
# ################################################################
#                       add diag geom
# ################################################################


def _plot_diag_geom(
    dax=None,
    key_cam=None,
    dplot=None,
    is2d=None,
):

    for k0, v0 in dplot.items():

        for k1, v1 in v0.items():

            # cross
            kax = 'cross'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['r'],
                        v1['z'],
                        v1['ur'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['r'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # hor
            kax = 'hor'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['ux'],
                        v1['uy'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        **v1.get('props', {}),
                    )

            # 3d
            kax = '3d'
            if dax.get(kax) is not None:
                ax = dax[kax]['handle']

                if k1.startswith('v-'):
                    ax.quiver(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        v1['ux'],
                        v1['uy'],
                        v1['uz'],
                        **v1.get('props', {}),
                    )

                else:
                    ax.plot(
                        v1['x'],
                        v1['y'],
                        v1['z'],
                        **v1.get('props', {}),
                    )

            # plotting of 2d camera contour
            kax = f"{k0}_sig"
            if is2d and k0 in key_cam and dax.get(kax) is not None:
                ax = dax[kax]['handle']
                if k1 == 'o':
                    ax.plot(
                        v1['x0'],
                        v1['x1'],
                        **v1.get('props', {}),
                    )


# ################################################################
# ################################################################
#                       add mobile
# ################################################################


def _add_camera_los_cross(
    coll2=None,
    k0=None,
    ax=None,
    kax=None,
    nlos=None,
    dref_los=None,
    dref_vos=None,
    color_dict=None,
    nan_los=None,
    nan_vos=None,
    alpha=None,
):

    for ii in range(nlos):

        # ------
        # los

        if nan_los is not None:
            l0, = ax.plot(
                nan_los,
                nan_los,
                c=color_dict['x'][ii%len(color_dict['x'])],
                ls='-',
                lw=1.,
            )

            # add mobile
            kl0 = f'{k0}_los_cross{ii}'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=dref_los[k0],
                data=[f'{k0}_los_r', f'{k0}_los_z'],
                dtype=['xdata', 'ydata'],
                axes=kax,
                ind=ii,
            )

        # ------
        # vos

        if nan_vos is not None:
            l0, = ax.fill(
                nan_vos,
                nan_vos,
                fc=color_dict['x'][ii%len(color_dict['x'])],
                alpha=alpha,
                ls='None',
                lw=0.,
            )

            # add mobile
            kl0 = f'{k0}_vos_cross{ii}'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=dref_vos[k0],
                data=[f'{k0}_vos_cross'],
                dtype=['xy'],
                axes=kax,
                ind=ii,
            )


def _add_camera_los_hor(
    coll2=None,
    k0=None,
    ax=None,
    kax=None,
    nlos=None,
    dref_los=None,
    dref_vos=None,
    color_dict=None,
    nan_los=None,
    nan_vos=None,
    alpha=None,
):

    for ii in range(nlos):

        # ------
        # los

        if nan_los is not None:
            l0, = ax.plot(
                nan_los,
                nan_los,
                c=color_dict['x'][ii%len(color_dict['x'])],
                ls='-',
                lw=1.,
            )

            # add mobile
            kl0 = f'{k0}_los_hor{ii}'
            coll2.add_mobile(
                key=kl0,
                handle=l0,
                refs=dref_los[k0],
                data=[f'{k0}_los_x', f'{k0}_los_y'],
                dtype=['xdata', 'ydata'],
                axes=kax,
                ind=ii,
            )

        # ------
        # vos

        if f'{k0}_vos_hor' in coll2.ddata.keys():

            if nan_vos is not None:
                l0, = ax.fill(
                    nan_vos,
                    nan_vos,
                    fc=color_dict['x'][ii%len(color_dict['x'])],
                    alpha=alpha,
                    ls='None',
                    lw=0.,
                )

                # add mobile
                kl0 = f'{k0}_vos_hor{ii}'
                coll2.add_mobile(
                    key=kl0,
                    handle=l0,
                    refs=dref_vos[k0],
                    data=[f'{k0}_vos_hor'],
                    dtype=['xy'],
                    axes=kax,
                    ind=ii,
                )


def _add_camera_vlines_marker(
    coll2=None,
    dax=None,
    ax=None,
    kax=None,
    is2d=None,
    k0=None,
    nlos=None,
    ddatax=None,
    ddatay=None,
    drefx=None,
    drefy=None,
    dkeyx=None,
    dkeyy=None,
    color_dict=None,
    suffix=None,
):

    ncolx = len(color_dict['x'])
    ncoly = len(color_dict['y'])
    if suffix is None:
        suffix = ''

    if is2d:
        for ii in range(nlos):
            mi, = ax.plot(
                ddatax[k0][0:1],
                ddatay[k0][0:1],
                marker='s',
                ms=6,
                markeredgecolor=color_dict['x'][ii%ncolx],
                markerfacecolor='None',
            )

            km = f'{k0}_m{ii:02.0f}{suffix}'
            coll2.add_mobile(
                key=km,
                handle=mi,
                refs=[drefx[k0], drefy[k0]],
                data=[dkeyx[k0], dkeyy[k0]],
                dtype=['xdata', 'ydata'],
                axes=kax,
                ind=ii,
            )

        dax[kax].update(
            refx=[drefx[k0]],
            refy=[drefy[k0]],
            datax=[dkeyx[k0]],
            datay=[dkeyy[k0]],
        )

    else:

        for ii in range(nlos):
            lv = ax.axvline(
                ddatax[k0][0], c=color_dict['y'][ii%ncoly], lw=1., ls='-',
            )
            kv = f'{k0}_v{ii:02.0f}{suffix}'
            coll2.add_mobile(
                key=kv,
                handle=lv,
                refs=drefx[k0],
                data=dkeyx[k0],
                dtype='xdata',
                axes=kax,
                ind=ii,
            )

        dax[kax].update(refx=[drefx[k0]], datax=[dkeyx[k0]])


"""
This module contains tests for tofu.geom in its structured version
"""

# External modules
import sys
import os
import numpy as np
import matplotlib.pyplot as plt


_PATH_HERE = os.path.dirname(__file__)
_PATH_TOFU = os.path.dirname(os.path.dirname(os.path.dirname(_PATH_HERE)))


sys.path.insert(0, _PATH_TOFU)
import tofu as tf
sys.path.pop(0)


#######################################################
#
#     Setup and Teardown
#
#######################################################

def setup_module(module):
    print ("") # this is to get a newline after the dots
    #print ("setup_module before anything in this file")

def teardown_module(module):
    #os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    #os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    #print ("teardown_module after everything in this file")
    #print ("") # this is to get a newline
    pass


#######################################################
#
#     Utility setup routines
#
#######################################################


def _get_cases(ang0=np.pi/4., ang1=-np.pi/4.):

    # load and sample config
    conf = tf.load_config('WEST-V0')
    pts, dV, ind, reseff = conf.Ves.V1.get_sampleV(
        res=0.1,
        domain=[[2.2, 2.6], [-0.1, 0.1], [-np.pi/10, np.pi/10]],
    )

    # Main line
    P0 = np.r_[2.5, 0, 0]
    vect = np.array([
        np.cos(ang0)*np.cos(ang1),
        np.cos(ang0)*np.sin(ang1),
        np.sin(ang0),
    ])
    e0 = np.array([-vect[1], vect[0], 0])
    e0 /= np.linalg.norm(e0)
    e1 = np.cross(e0, vect)

    # centers positions in X, Y, Z
    nsides = np.r_[3, 4, 6, 4, 4]
    dx1 = np.r_[0, 0, 0, 0, 0.05]
    radius = 0.02
    lengths = 1. + np.linspace(0, 1, len(nsides))
    cents = (
        P0[:, None]
        + lengths[None, :]*vect[:, None]
        + dx1[None, :]*e1[:, None]
    )

    # outlines (triangle, rectangle, hexagon...)
    lout = []
    for ii, nn in enumerate(nsides):
        theta = np.linspace(-np.pi, np.pi, nn+1)
        lout.append(radius * np.array([np.cos(theta), np.sin(theta)]))

    # dict
    lcases = [f'case {ii}' for ii in range(2)]  # 4
    dcases = dict.fromkeys(lcases, {})
    for ii, cc in enumerate(lcases):
        dcases[cc]['pts_x'] = pts[0, :]
        dcases[cc]['pts_y'] = pts[1, :]
        dcases[cc]['pts_z'] = pts[2, :]
        dcases[cc]['visibility'] = ii % 2 == 1
        dcases[cc]['return_vector'] = ii % 4 < 2
        dcases[cc]['apertures'] = {
            'ap{jj}': {
                'poly_x': (
                    cents[0, ii]
                    + e0[0] * lout[jj][0, :]
                    + e1[0] * lout[jj][1, :]
                ),
                'poly_y': (
                    cents[1, ii]
                    + e0[1] * lout[jj][0, :]
                    + e1[1] * lout[jj][1, :]
                ),
                'poly_z': (
                    cents[2, ii]
                    + e0[2] * lout[jj][0, :]
                    + e1[2] * lout[jj][1, :]
                ),
                'nin': vect,
            }
            for jj in range(3)
        }
        dcases[cc]['detectors'] = {
            'outline_x0': lout[-1][0, :],
            'outline_x1': lout[-1][1, :],
            'cents_x': cents[0, -2:],
            'cents_y': cents[1, -2:],
            'cents_z': cents[2, -2:],
            'nin_x': -vect[0]*np.r_[1, 1],
            'nin_y': -vect[1]*np.r_[1, 1],
            'nin_z': -vect[2]*np.r_[1, 1],
            'e0_x': e0[0]*np.r_[1, 1],
            'e0_y': e0[1]*np.r_[1, 1],
            'e0_z': e0[2]*np.r_[1, 1],
            'e1_x': e1[0]*np.r_[1, 1],
            'e1_y': e1[1]*np.r_[1, 1],
            'e1_z': e1[2]*np.r_[1, 1],
        }
        dcases[cc]['config'] = conf if dcases[cc]['visibility'] else None

    return dcases


#######################################################
#
#     Class
#
#######################################################


class Test01_SolidAngles():

    # ------------------------
    #   setup
    # ------------------------

    def setup_method(self):

        self.dcases = _get_cases()

    # ------------------------
    #   Populating
    # ------------------------

    def test01_solid_angle_apertures(self):

        for k0, v0 in self.dcases.items():

            nd = v0['detectors']['cents_x'].size
            na = len(v0['apertures'])
            npts = v0['pts_x'].size

            # call function
            out = tf.geom._comp_solidangles.calc_solidangle_apertures(
                # observation points
                pts_x=v0['pts_x'],
                pts_y=v0['pts_y'],
                pts_z=v0['pts_z'],
                # polygons
                apertures=v0['apertures'],
                detectors=v0['detectors'],
                # possible obstacles
                config=v0.get('config'),
                # parameters
                visibility=v0['visibility'],
                return_vector=v0['return_vector'],
            )

            # prepare checks
            if not v0['return_vector']:
                out = (out,)
                nout = 1
            else:
                nout = 4

            # Check output format conformity
            assert isinstance(out, tuple) and len(out) == nout
            assert all([
                isinstance(oo, np.ndarray)
                and oo.shape == (nd, npts)
                for oo in out
            ]), f'{(nd, npts)} vs {[oo.shape for oo in out]}'
            assert np.all(out[0] >= 0.)

            # check unit vectors (normalized)
            if nout > 1:
                iok = (out[0] > 0)
                assert all([
                    np.all(np.isfinite(oo) == iok)
                    for oo in out[1:]
                ])
                norm2 = out[1]**2 + out[2]**2 + out[3]**2
                assert np.allclose(norm2[iok], 1)



# distutils: language=c++
# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions for the earclipping techinque for triangulation of a
# polygon. This is really useful when vignetting, as the computation to know
# if a ray intersected a polygon, is easier if we check if the polygon is
# discretized in triangles and then we check if the ray intersected each
# triangle.
################################################################################

from libc.stdint cimport int64_t

cdef void compute_diff2d(
    double* orig,
    int nvert,
    double* diff,
) nogil

# DEPRECATED ?
cdef void compute_diff3d(double* orig,
                         int nvert,
                         double* diff) nogil

cdef bint is_reflex_2d(
    const double[1] u0,
    const double[1] u1,
    const double[1] v0,
    const double[1] v1,
) nogil

# DEPRECATED
cdef bint is_reflex_3d(
    const double[3] u,
    const double[3] v,
    double[3] vect_cc
) nogil

cdef void are_points_reflex_2d(
    int nvert,
    double* diff,
    bint* are_reflex,
) nogil

# DEPRECATED ?
cdef  void are_points_reflex_3d(int nvert,
                                double* diff,
                                bint* are_reflex,
                                double[3] vect_cc) nogil

cdef bint is_pt_in_tri_2d(
    double Ax,
    double Ay,
    double Bx,
    double By,
    double Cx,
    double Cy,
    double px,
    double py,
) nogil

# DEPRECATED
cdef bint is_pt_in_tri_3d(
    double[3] v0, double[3] v1,
    double Ax, double Ay, double Az,
    double px, double py, double pz,
) nogil

# cdef int get_one_ear(
    # double* polygon,
    # bint* lref,
    # _cl.ChainedList* working_index,
    # int nv,
    # int nvert,
# ) nogil

cdef void earclipping_poly_2d(
    double* vignett,
    int64_t* ltri,
    double* diff,
    bint* lref,
    int nvert,
) nogil

cdef void triangulate_poly(double* vignett_poly,
                          int64_t nvert,
                          int64_t** ltri) nogil

cdef int triangulate_polys(double** vignett_poly,
                            # int64_t* lnvert,
                            int64_t* lnvert,
                            int nvign,
                            int64_t** ltri,
                            int num_threads) except -1 nogil


# ===============================================================
#               Vignetting
# ===============================================================


cdef bint inter_ray_poly(
    const double[3] ray_orig,
    const double[3] ray_vdir,
    double* vignett,
    int nvert,
    int64_t* ltri,
) nogil

cdef void vignetting_core(double[:, ::1] ray_orig,
                          double[:, ::1] ray_vdir,
                          double** vignett,
                          # int64_t* lnvert,
                          int64_t* lnvert,
                          double* lbounds,
                          int64_t** ltri,
                          int nvign,
                          int nlos,
                          bint* goes_through,
                          int num_threads) nogil

cdef int vignetting_vmesh_vpoly(int npts, int sz_r,
                                bint is_cart,
                                double[:, ::1] vpoly,
                                double[:, ::1] pts,
                                double[::1] vol_resol,
                                double[::1] r_on_phi,
                                double* disc_r,
                                int64_t[::1] lind,
                                double** res_x,
                                double** res_y,
                                double** res_z,
                                double** res_vres,
                                double** res_rphi,
                                int64_t** res_lind,
                                int64_t* sz_rphi,
                                int num_threads) nogil

cdef int are_in_vignette(int sz_r, int sz_z,
                         double[:, ::1] vpoly,
                         int npts_vpoly,
                         double* disc_r,
                         double* disc_z,
                         int64_t[:, ::1] is_in_vignette) nogil




import numpy as np
try:
    from sksparse.cholmod import cholesky
except Exception:
    cholesky = False


import tomotok.core as tmtkc


# #############################################################################
# #############################################################################
#           Define DALGO for tomotok
# #############################################################################


def _get_sparse_from_docstr(doc, k0=None):
    gmat = [ss for ss in doc if ss.strip().startswith('gmat : ')]
    if len(gmat) != 1:
        msg = (
            f"tomotok class {k0}\n"
            "No/too many matches found in docstr for 'gmat : '\n"
            f"\t- matches: {gmat}"
        )
        raise Exception(msg)
    if any([ss in gmat[0] for ss in ['scipy', 'sparse', 'csr_matrix']]):
        sparse = True
    elif any([ss in gmat[0] for ss in ['np.ndarray', 'numpy']]):
        sparse = False
    else:
        msg = (
            f"tomotok class {k0}\n"
            f"Unreckognized type of gmat: {gmat}"
        )
        raise Exception(msg)
    return sparse


def get_dalgo():

    # ----------------
    # Initialize dict

    dalgo = dict.fromkeys([
        ss for ss in dir(tmtkc.inversions)
        if type(getattr(tmtkc.inversions, ss)) == type
        and not (cholesky is False and ss == 'CholmodMfr')
        and 'bob' not in ss.lower()
    ])

    # ---------
    # Complete

    for k0 in dalgo.keys():

        # family
        fam = 'Non-regularized' if 'bob' in k0.lower() else 'Phillips-Tikhonov'

        # reg. operator
        if 'mfr' in k0.lower():
            regoper = 'MinFisher'
        elif 'algebraic' in k0.lower():
            regoper = 'any linear'
        elif 'bob' in k0.lower():
            regoper = False

        # reg. parameter
        if 'algebraic' in k0.lower():
            regparam = 'linear estimate'
        else:
            regparam = ''

        # matrix decomposition
        if 'chol' in k0.lower():
            decomp = 'cholesky'
        elif 'svd' in k0.lower():
            decomp = 'svd'
        elif 'Gev' in k0:
            decomp = 'Gen. Eigen val.'
        else:
            decomp = ''

        # positivity constraint
        pos = False

        # sparse
        doc = getattr(tmtkc.inversions, k0).__call__.__doc__.split('\n')
        sparse = _get_sparse_from_docstr(doc, k0=k0)

        # fill
        dalgo[k0] = {
            'source': 'tomotok',
            'family': fam,
            'reg_operator': regoper,
            'reg_param': regparam,
            'decomposition': decomp,
            'positive': pos,
            'sparse': sparse,
            'isotropic': True,
            'func': k0,
        }

    return dalgo


# #############################################################################
# #############################################################################
#           Define functions to be called by tofu
# #############################################################################


def SvdFastAlgebraic(
    sig_norm=None,
    gmat_norm=None,
    deriv=None,
    method=None,
    num=None,
    # additional
    nchan=None,
    **kwdargs,
):

    # solve
    solver = tmtkc.SvdFastAlgebraic()
    sol = solver.invert(
        sig_norm,
        gmat_norm,
        deriv,
        method=None,
        num=None,
    )

    # compute residue
    chi2n = np.sum((gmat_norm.dot(sol) - sig_norm)**2) / nchan
    reg = sol.dot(deriv.dot(sol))

    return sol, chi2n, reg


def GevFastAlgebraic(
    sig_norm=None,
    gmat_norm=None,
    deriv=None,
    method=None,
    num=None,
    # additional
    nchan=None,
    **kwdargs,
):
    """
    Expects deriv to be provided as tgradR.dot(gradR) + tgradZ.dot(gradZ)

    """

    # solve
    solver = tmtkc.GevFastAlgebraic()
    sol = solver.invert(
        sig_norm,
        gmat_norm,
        deriv,
        method=None,
        num=None,
    )

    # compute residue
    chi2n = np.sum((gmat_norm.dot(sol) - sig_norm)**2) / nchan
    reg = sol.dot(deriv.dot(sol))

    return sol, chi2n, reg


def Mfr(
    sig_norm=None,
    gmat_norm=None,
    deriv=None,
    method=None,
    num=None,
    # additional
    nchan=None,
    **kwdargs,
):
    """
    Expects deriv to be provided as a tuple (gradR, gradZ)

    """

    # solve
    solver = tmtkc.Mfr()
    sol = solver.invert(
        sig_norm,
        gmat_norm,
        deriv,          # here must be provided as (gradR, gradZ)
        w_factor=None,
        mfi_num=3,
        bounds=(-15, 0),
        iter_max=10,
        w_max=1,
        danis=0,
    )

    # compute residue
    chi2n = np.sum((gmat_norm.dot(sol) - sig_norm)**2) / nchan
    reg = sol.dot(deriv[0][0].dot(sol)) + sol.dot(deriv[0][1].dot(sol))

    return sol, chi2n, reg


def CholmodMfr(
    sig_norm=None,
    gmat_norm=None,
    deriv=None,
    method=None,
    num=None,
    # additional
    nchan=None,
    **kwdargs,
):
    """
    Expects deriv to be provided as a tuple (gradR, gradZ)

    """

    # solve
    solver = tmtkc.CholmodMfr()
    sol = solver.invert(
        sig_norm,
        gmat_norm,
        deriv,          # here must be provided as (gradR, gradZ)
        w_factor=None,
        mfi_num=3,
        bounds=(-15, 0),
        iter_max=10,
        w_max=1,
        danis=0,
    )

    # compute residue
    chi2n = np.sum((gmat_norm.dot(sol) - sig_norm)**2) / nchan
    reg = sol.dot(deriv[0][0].dot(sol)) + sol.dot(deriv[0][1].dot(sol))

    return sol, chi2n, reg







import numpy as np
import matplotlib.pyplot as plt

# ToFu-specific
import tofu.geom as tfg
try:
    import tofu.dust._comp as _comp
    import tofu.dust._plot as _plot
except Exception:
    from . import _comp as _comp
    from . import _plot as _plot






class Dust(object):
    """ A generic Dust class to handle dust trajectories, radiative power
    balance...

    """

    def __init__(self, Id=None, dmat=None, traj=None, VType='Tor',
                 Ves=None, LStruct=None, res=None):
        # Initialize fields
        self._Id = None
        self._dmat = None
        self._traj = None
        self._geom = None
        self._res = None
        self._emiss = None
        self._direct = None
        if Id is not None:
            self._set_Id(Id)
        if dmat is not None:
            self.set_material_properties(dmat)
        if traj is not None:
            self.set_traj(traj)
        if Ves is not None:
            self.set_geom(Ves=Ves, LStruct=LStruct, res=res)

    @property
    def dmat(self):
        return self._dmat
    @property
    def traj(self):
        return self._traj
    @property
    def Ves(self):
        if self._geom is None:
            out = None
        else:
            out = self._geom['Ves']
        return out
    @property
    def LStruct(self):
        if self._geom is None:
            out = None
        else:
            out = self._geom['LStruct']
        return out
    @property
    def res(self):
        return self._res
    @property
    def emiss(self):
        return self._emiss
    @property
    def direct(self):
        return self._direct

    def set_material_properties(self, dmat=None):
        assert dmat is None or type(dmat) is dict
        self._dmat = dmat

    def set_traj(self, traj=None):
        C0 = traj is None
        C1 = type(traj) is dict
        C2 = hasattr(traj,'__iter__')
        assert C0 or C1 or C2
        if C1:
            assert 'pts' in traj.keys()
            traj['pts'] = _check_trajpts(traj['pts'])
            if 't' in traj.keys():
                traj['t'] = _check_trajt(traj['t'],traj['pts'].shape[0])
        if C2:
            pts = _check_trajpts(traj)
            traj = {'pts':pts}
        traj['npts'] = traj['pts'].shape[0]
        self._traj = traj

    def set_geom(self, Ves=None, LStruct=None, res=None):
        msg = "Arg Ves must be a tf.geom.Ves"
        assert Ves is None or isinstance(Ves,tfg.Ves), msg
        C0 = LStruct is None
        C1 = isinstance(LStruct,tfg.Struct)
        C2 = (type(LStruct) is list
              and all([isinstance(ss,tfg.Struct) for ss in LStruct]))
        msg = ""
        assert C0 or C1 or C2, msg
        assert Ves is not None or LStruct is None, "Ves must be provided !"

        LStruct = [LStruct] if isinstance(LStruct, tfg.Struct) else LStruct
        self._geom = {'Ves':Ves, 'LStruct':LStruct}
        self._sampleV = None
        if res is not None:
            self.set_sampleV(res)

    def set_sampleV(self, res=None):
        #assert
        self._res = res

    def set_emiss(self, emiss=None, t=None, Ani=None, axisym=True):
        ani = tfg._GG.check_ff(emiss, t=t, Ani=Ani, Vuniq=False)
        self._emiss = {'ff':emiss, 't':t, 'ani':ani, 'axisym':axisym}

    def set_directproblem(self, pts=None, r=None, t=None):
        C0 = type(pts) is np.ndarray
        C1 = C0 and pts.ndim==2 and pts.shape[0]==3
        C2 = C0 and pts.shape==(3,)
        msg = "pts must be a (3,) or (3,N) np.ndarray (X,Y,Z coordinates of N points)"
        assert C1 or C2, msg
        if C2:
            pts = pts.reshape((3,1))
        nt = pts.shape[1]

        C0 = type(r) in [int, float, np.int64, np.float64]
        C1 = type(r) is np.ndarray and r.ndim==1
        C2 = C1 and r.size>1
        msg = "r must be a float or a 1D np.ndarray"
        assert C0 or C1, msg
        if C0:
            r = np.array([r],dtype=float)
        if r.size==1:
            r = np.tile(r,nt)
        assert r.size==nt, "r must be a (nt,) np.ndarray !"

        C0 = t is None or type(t) is np.ndarray and t.shape==(nt,)
        msg = "t must be None or a (N,) np.ndarray"
        assert C0, msg
        if t is None:
            t = np.arange(0,nt)

        self._direct = {'pts':pts, 'r':r, 't':t}


    """
    def calc_directproblem(self):
        msg = "set the direct problem before !"
        assert self.direct is not None, msg
        if self.emiss is None:
            msg = "emiss not set => only the solid angle will be computed !"
            warnings.warn(msg)
            out = _comp.
            lpolyCross, lpolyHor, gridCross, gridHor = out[:4]
            saCross, saHor, volIn, saIn = out[4:]
        else:
            #out =
            lpolyCross, lpolyHor, gridCross, gridHor = out[:4]
            saCross, saHor, volIn, saIn = out[4:8]
            contribCross, contribHor, powIn = out[8:]

        self._direct_sol = {'lpolyCross':lpolyCross, 'lpolyHor':lpolyHor,
                            'gridCross'}
    """

    def plot(self):
        dax, KH = _plot.plot(self)
        return dax, KH



def _check_trajpts(pts):
    pts = np.asarray(pts).astype(float)
    assert pts.ndim in [1,2]
    if pts.ndim==1:
        assert pts.size==3
        pts = pts.reshape((1,3))
    assert 3 in pts.shape
    if not pts.shape[1]==3:
        pts = pts.T
    return pts

def _check_trajt(t, npts):
    t = np.asarray(t).astype(float)
    assert t.ndim==1
    assert t.size==npts
    return t



# -*- coding: utf-8 -*-
"""
Created on Thu Dec 21 15:36:19 2023

@author: dvezinet
"""


import os


import numpy as np
import datastock as ds


# ###############################################################
# ###############################################################
#                           Main
# ###############################################################


def add_crystal(
    coll=None,
    cryst=None,
    key=None,
):

    # -------------------
    # check inputs
    # -------------------

    coll, cryst, key = _check(
        coll=coll,
        cryst=cryst,
        key=key,
    )

    # --------------------
    # extract geometry
    # --------------------

    dgeom = _extract_dgeom(cryst)

    # --------------------
    # extract material
    # --------------------

    dmat = _extract_dmat(cryst)

    # --------------------
    # add
    # --------------------

    coll.add_crystal(
        key=key,
        dgeom=dgeom,
        dmat=dmat,
        color=cryst._dmisc['color'],
    )

    return


# ###############################################################
# ###############################################################
#                           Check
# ###############################################################

def _check(
    coll=None,
    cryst=None,
    key=None,
):


    # -------------------
    # check coll
    # -------------------

    import tofu as tf
    if coll is None:
        coll = tf.data.Collection()

    if not isinstance(coll, tf.data.Collection):
        msg = (
            "Arg coll must be a tf.data.Collection instance!\n"
            "Provided:\n"
            "\t- class: {type(coll)}\n"
            "\t- value: {coll}\n"
        )
        raise Exception(msg)

    # -------------------
    # check crystal
    # -------------------

    # load from file
    if isinstance(cryst, str):
        error = False
        if not (os.path.isfile(cryst) and cryst.endswith('.npz')):
            error = 'not a valid file name'
        else:
            try:
                cryst = tf.load(cryst)
            except Exception as err:
                error = f'tf.load() failed to load {cryst}\n{str(err)}'

        # raise error
        if error is not False:
            msg = (
                f"Arg cryst must be a valid .npz file, loadable!\n{error}"
            )
            raise Exception(msg)

    # check class
    if not cryst.__class__.__name__ == 'CrystalBragg':
        msg = (
            "Arg cryst must be a legacy tf.geom.CrystalBragg instance!\n"
            "Provided:\n"
            f"\t- class: {type(cryst)}\n"
            f"\t- value: {cryst}\n"
        )
        raise Exception(msg)

    # -------------------
    # check key
    # -------------------

    # default value
    kdef = cryst.Id.Name

    # check
    lout = list(coll.dobj.get('crystal', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        default=kdef,
        excluded=lout,
    )

    return coll, cryst, key


# ###############################################################
# ###############################################################
#                       Extract dgeom
# ###############################################################


def _extract_dgeom(cryst):

    # ----------------------------
    # intialize with unit vectors

    dgeom = {
        'nin': np.copy(cryst._dgeom['nin']),
        'e0': np.copy(-cryst._dgeom['e1']),
        'e1': np.copy(cryst._dgeom['e2']),
    }

    # ----------
    # get center

    dgeom['cent'] = np.copy(cryst._dgeom['summit'])

    # ---------------
    # get extenthalf

    dgeom['extenthalf'] = np.copy(cryst._dgeom['extenthalf'])

    # -----------------------
    # get radius of curvature

    typ = cryst.dgeom['Type']
    rc = float(cryst._dgeom['rcurve'])
    if typ == 'sph':
        dgeom['curve_r'] = np.r_[rc, rc]
    else:
        msg = (
            "Suspicious, legacy tofu only handles spherical CrystalBragg!\n"
            f"Provided:\n\t- 'Type': {typ}"
        )
        raise NotImplementedError(msg)

    return dgeom


# ###############################################################
# ###############################################################
#                       Extract dmat
# ###############################################################


def _extract_dmat(cryst):

    # ------------
    # dmat

    dmat = {
        'material': str(cryst._dmat['formula']),
        'name': None,
        'symbol': None,
        'mesh': {
            'type': cryst._dmat['symmetry'],
        },
        'target': {
            'lamb': float(cryst._dbragg['lambref']),
            'bragg': float(cryst._dbragg['braggref']),
        },
        'd_hkl': float(cryst._dmat['d']),
        'alpha': float(cryst._dmat['alpha']),
        'beta': float(cryst._dmat['beta']),
        'density': float(cryst._dmat['density']),
        'miller': np.r_[cryst._dmat['cut'][:2], cryst._dmat['cut'][-1]],
    }

    # -------------
    # rocking curve

    if cryst._dbragg.get('rockingcurve') is not None:
        dmat['drock'] = {
            'angle_rel': cryst._dbragg['rockingcurve']['dangle'],
            'power_ratio': cryst._dbragg['rockingcurve']['value'],
        }

    return dmat


# -*- coding: utf-8 -*-


import os


import numpy as np
import scipy.interpolate as scpinterp
import datastock as ds


# #################
# #################
# DEFAULTS
# #################


# PATH
_PATH_HERE = os.path.dirname(__file__)
_PATH_INPUTS = os.path.join(_PATH_HERE, 'inputs_filter')


# MASS DENSITIE, g/cm3
_DMASS_DENS = {
    'LaBr3': 5.06,
}


# AVAILABLE MATERIALS
_LMAT_LENGTH = [
    ss for ss in os.listdir(_PATH_INPUTS)
    if ss.startswith('AttenuationLength_')
    and ss.endswith('.txt')
    and ss.count('_') == 1
]
_LMAT_LENGTH = [ss.split('_')[1].replace('.txt', '') for ss in _LMAT_LENGTH]
_LMAT_MASS = [
    ss for ss in os.listdir(_PATH_INPUTS)
    if ss.endswith('_MassAttenuationCoef.csv')
    and ss.count('_') == 1
]
_LMAT_MASS = [ss.split('_')[0] for ss in _LMAT_MASS]
_LMAT_MASS = [ss for ss in _LMAT_MASS if ss in _DMASS_DENS.keys()]
_LOK_MAT = _LMAT_LENGTH + _LMAT_MASS


# #################################################################
# #################################################################
#               main
# #################################################################


def main(
    dthick=None,
    E=None,
    # plotting
    plot=None,
):
    """ Return dict of linear transmission

    dthick is a dict of:
        - 'key': {
            'mat': str (refers to available),
            'thick': thickness (scalar or array, in meters),
        }

    Several different keys can use the same material if needs be

    If plot=True, 'key' will be used for labelling

    E = array of energy values in eV

    """

    # ---------------
    # check inputs
    # ---------------

    dthick, E, plot = _check(
        dthick=dthick,
        E=E,
        plot=plot,
    )

    # -----------------
    # initialize output
    # -----------------

    dout = {
        'E': {
            'data': E,
            'units': 'eV',
        },
        'keys': {},
        'total': {},
    }

    # ---------------
    # loop on keys
    # ---------------

    transtot = None
    for k0, v0 in dthick.items():

        # get transmission function from type
        if v0['type'] == 'length':
            func = _get_transmission_from_length
        else:
            func = _get_transmission_from_mass

        # compute
        trans = func(
            mat=v0['mat'],
            E=E,
            thick=v0['thick'],
        )

        # store
        dout['keys'][k0] = {
            'trans': trans,
            'thick': v0['thick'],
        }

        # transtot
        if transtot is None:
            transtot = trans
        else:
            transtot = transtot * trans

    # -------------------
    # total transmission
    # -------------------

    dout['total'] = transtot

    # --------------------
    # plot
    # --------------------

    if plot is True:
        pass
        # fig = plt.figure()

        # ax = fig.add_subplot()

        # for kloc in  dfilter[boxi].keys():
        # ax.semilogy(
        # E,
        # dfilter[boxi][kloc]['total'],
        # ls='-',
        # lw=1,
        # label=kloc,
        # )
        # ax.legend()

    return dout


# ###############################################################################
# ###############################################################################
#                Check inputs
# ###############################################################################


def _check(
    dthick=None,
    E=None,
    plot=None,
):

    # -----------------
    # dthick
    # -----------------

    # -----------------
    # first check

    c0 = (
        isinstance(dthick, dict)
        and all([
            isinstance(k0, str)
            and isinstance(v0, dict)
            for k0, v0 in dthick.items()
        ])
    )
    if not c0:
        _error_dthick(dthick)

    # -----------------
    # second check

    dfail = {}
    for k0, v0 in dthick.items():

        # material
        mat = v0.get('mat')
        if mat not in _LOK_MAT:
            dfail[k0] = f"'mat' {mat} not available"
            continue

        # type
        if mat in _LMAT_LENGTH:
            dthick[k0]['type'] = 'length'
        else:
            dthick[k0]['type'] = 'mass'

        # thickness
        thick = v0.get('thick')
        try:
            thick = np.atleast_1d(thick).astype(float)
            assert np.all(np.isfinite(thick))
            assert np.all(thick > 0.)
            dthick[k0]['thick'] = thick
        except Exception as err:
            dfail[k0] = (
                "'thick' must be convertible to a strictly"
                " positive finite array of floats"
            )
            raise err

    # ----------------------
    # Raise Exception if any

    if len(dfail) > 0:
        _error_dthick(dthick, dfail=dfail)

    # -----------------
    # E
    # -----------------

    # ---------
    # E itself

    try:
        E = np.asarray(E).astype(float)
        assert np.all(np.isfinite(E))
        assert np.all(E > 0)
    except Exception as err:
        msg = (
            "Arg E must be convertible to a np.ndarray "
            "of strictly positive float\n"
            f"Provided:\n{E}\n"
        )
        raise Exception(msg) from err

    # ------------------
    # Broadcastability
    # -------------------

    dthick, shapef = ds._generic_check._check_all_broadcastable(
        E=E,
        **dthick,
    )

    E = dthick['E']
    dthick = {k0: v0 for k0, v0 in dthick.items() if k0 != 'E'}

    # -----------------
    # plot
    # -----------------

    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    return dthick, E, plot


# #################
# Exception
# #################


def _error_dthick(dthick, dfail=None):

    msg = (
        "Arg dthick must be a dict with a series of:\n"
        "\t- 'key0': {'mat': str, 'thick': float or array},\n"
        "\t- 'key1': {'mat': str, 'thick': float or array},\n"
        "\t- ...\n"
        "\t- 'key2N': {'mat': str, 'thick': float or array},\n"
        "Where 'mat' must refer to an available material, in:\n"
        f"{_LOK_MAT}\n\n"
    )

    if dfail is None:
        msg += (
            "Provided:\n"
            f"{dthick}"
        )

    else:
        lstr = [f"\t- {k0}: {v0}" for k0, v0 in dfail.items()]
        msg += (
            "\n" + "\n".join(lstr)
        )

    raise Exception(msg)


# ###############################################################################
# ###############################################################################
#               get transmission
# ###############################################################################


def _get_transmission_from_length(mat=None, E=None, thick=None):

    # load absorption length
    name = f"AttenuationLength_{mat}.txt"
    pfe = os.path.join(_PATH_INPUTS, name)
    # length in microns
    Att_E, Att_L = np.loadtxt(pfe).T
    # length microns => m
    Att_L = Att_L * 1e-6

    # compute transmision
    length = 10**(scpinterp.interp1d(
        np.log10(Att_E),
        np.log10(Att_L),
        kind='linear',
        bounds_error=False,
        fill_value=(np.log10(Att_L[0]), np.log10(Att_L[-1])),
    )(np.log10(E)))

    return np.exp(-thick / length)


def _get_transmission_from_mass(mat=None, E=None, thick=None):

    """

    absorp = 1 - np.exp(
        - 0.02 / scpinterp.interp1d(
            E_abs * 1e6,
            att_length,
            kind='linear',
            bounds_error=False,
            fill_value=(att_length[0], att_length[-1]),
        )(E)
    )
    """

    # crystal absorption (MeV, cm2/g)
    name = f"{mat}_MassAttenuationCoef.csv"
    pfe = os.path.join(_PATH_INPUTS, name)
    E_abs, coef_mass = np.loadtxt(pfe, delimiter=',').T

    # intermediates
    mass_density = _DMASS_DENS[mat]
    # length, m
    att_length = 1e-2 / (coef_mass * mass_density)

    # interpolat log log
    length = 10**(scpinterp.interp1d(
        np.log10(E_abs) + 6.,  # MeV => eV
        np.log10(att_length),
        kind='linear',
        bounds_error=False,
        fill_value=(np.log10(att_length[0]), np.log10(att_length[-1])),
    )(np.log10(E)))

    trans = np.exp(-thick / length)

    return trans





# Built-in
import os
import itertools as itt

# Common
import numpy as np
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches
import matplotlib.colors as mcolors
from matplotlib.colors import ListedColormap
import matplotlib.gridspec as gridspec
from matplotlib.axes._axes import Axes
from mpl_toolkits.mplot3d import Axes3D

# tofu
from tofu.version import __version__

__all__ = [
    'plot_fit1d',
    'plot_dinput2d',
    'plot_fit2d',
]


_GITHUB = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at %s' % (
                __version__, _GITHUB
                )

_QUIVERCOLOR = plt.cm.viridis(np.linspace(0, 1, 3))
_QUIVERCOLOR = np.array([[1., 0., 0., 1.],
                         [0., 1., 0., 1.],
                         [0., 0., 1., 1.]])
_QUIVERCOLOR = ListedColormap(_QUIVERCOLOR)


# #################################################################
# #################################################################
#                   data plot
# #################################################################
# #################################################################


def CrystalBragg_plot_data_vs_lambphi(
    xi, xj, bragg, lamb, phi, data,
    lambfit=None, phifit=None,
    spect1d=None, vertsum1d=None,
    lambax=None, phiax=None,
    phiminmax=None, dlines=None,
    lambmin=None, lambmax=None,
    xjcut=None, lambcut=None,
    phicut=None, spectcut=None,
    cmap=None, vmin=None, vmax=None,
    fs=None, dmargin=None,
    tit=None, wintit=None,
    angunits='deg',
):

    # Check inputs
    # ------------

    if fs is None:
        fs = (14, 8)
    if tit is None:
        tit = False
    if wintit is None:
        wintit = _WINTIT
    if cmap is None:
        cmap = plt.cm.viridis
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.99,
                   'bottom': 0.07, 'top': 0.92,
                   'wspace': 0.2, 'hspace': 0.3}
    assert angunits in ['deg', 'rad']
    if angunits == 'deg':
        bragg = bragg*180./np.pi
        phi = phi*180./np.pi
        phifit = phifit*180./np.pi
        if phiax is not None:
            phiax = 180*phiax/np.pi
        phiminmax = phiminmax*180./np.pi
        if phicut is not None:
            phicut = phicut*180./np.pi
    nspect = spect1d.shape[0]

    if dlines is not None:
        lines = [k0 for k0, v0 in dlines.items()
                 if (v0['lambda0'] >= lambfit[0]
                     and v0['lambda0'] <= lambfit[-1])]
        lions = sorted(set([dlines[k0]['ION'] for k0 in lines]))
        nions = len(lions)
        dions = {k0: [k1 for k1 in lines if dlines[k1]['ION'] == k0]
                 for k0 in lions}
        dions = {k0: {'lamb': np.array([dlines[k1]['lambda0']
                                        for k1 in dions[k0]]),
                      'symbol': [dlines[k1]['symbol'] for k1 in dions[k0]]}
                 for k0 in lions}
        lcol = ['k', 'r', 'b', 'g', 'm', 'c']
        ncol = len(lcol)
        llamb = np.concatenate([dions[lions[ii]]['lamb']
                                for ii in range(nions)])
        indsort = np.argsort(llamb)
        lsymb = list(itt.chain.from_iterable([dions[lions[ii]]['symbol']
                                              for ii in range(nions)]))
        lsymb = [lsymb[ii] for ii in indsort]
        lcollamb = list(itt.chain.from_iterable(
            [[lcol[ii]]*dions[lions[ii]]['lamb'].size
             for ii in range(nions)]))
        lcollamb = [lcollamb[ii] for ii in indsort]

    lcolspect = ['k', 'r', 'b', 'g', 'y', 'm', 'c']
    ncolspect = len(lcolspect)

    # pre-compute
    # ------------

    # extent
    extent = (xi.min(), xi.max(), xj.min(), xj.max())
    if lambmin is None:
        lambmin = lambfit.min()
    if lambmax is None:
        lambmax = lambfit.max()

    # Plot
    # ------------

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(4, 4, **dmargin)
    ax0 = fig.add_subplot(gs[:3, 0], aspect='equal')   # , adjustable='datalim'
    ax1 = fig.add_subplot(
        gs[:3, 1],
        aspect='equal',
        sharex=ax0, sharey=ax0,
    )    # , adjustable='datalim'
    axs1 = fig.add_subplot(gs[3, 1], sharex=ax0)
    ax2 = fig.add_subplot(gs[:3, 2])
    axs2 = fig.add_subplot(gs[3, 2], sharex=ax2, sharey=axs1)
    ax3 = fig.add_subplot(gs[:3, 3], sharey=ax2)

    ax0.set_title('Coordinates transform')
    ax1.set_title('Camera image')
    axs1.set_title('Vertical average')
    ax2.set_title('Camera image transformed')

    ax0.set_ylabel(r'$x_j$ (m)')
    ax0.set_xlabel(r'$x_i$ (m)')
    axs1.set_ylabel(r'data')
    axs1.set_xlabel(r'$x_i$ (m)')
    ax2.set_ylabel(r'incidence angle ($deg$)')
    axs2.set_xlabel(r'$\lambda$ ($m$)')

    ax0.contour(xi, xj, bragg, 10, cmap=cmap)
    ax0.contour(xi, xj, phi, 10, cmap=cmap, ls='--')
    ax1.imshow(data, extent=extent, aspect='equal',
               origin='lower', vmin=vmin, vmax=vmax)
    axs1.plot(xi, np.nanmean(data, axis=0), c='k', ls='-')
    ax2.scatter(lamb.ravel(), phi.ravel(), c=data.ravel(), s=2,
                marker='s', edgecolors='None',
                cmap=cmap, vmin=vmin, vmax=vmax)

    if xjcut is None:
        ax3.plot(vertsum1d, phifit, c='k', ls='-')
    else:
        dphicut = 0.1*(phifit.max() - phifit.min())
        for ii in range(xjcut.size):
            ax1.axhline(xjcut[ii],
                        ls='-', lw=1., c='r')
            ax2.plot(lambcut[ii, :], phicut[ii, :],
                     ls='-', lw=1., c='r')
            ax3.plot(lambcut[ii, :],
                     (np.nanmean(phicut[ii, :])
                      + spectcut[ii, :]*dphicut/np.nanmax(spectcut[ii, :])),
                     ls='-', lw=1., c='r')

    if phiax is not None:
        ax2.plot(lambax, phiax, c='r', ls='-', lw=1.)

    # Plot known spectral lines
    if dlines is not None:
        for ii, k0 in enumerate(lions):
            for jj in range(dions[k0]['lamb'].size):
                x = dions[k0]['lamb'][jj]
                col = lcol[ii % ncol]
                axs2.axvline(x,
                             c=col, ls='--')
                axs2.annotate(dions[k0]['symbol'][jj],
                              xy=(x, 1.01), xytext=None,
                              xycoords=('data', 'axes fraction'),
                              color=col, arrowprops=None,
                              horizontalalignment='center',
                              verticalalignment='bottom')
                if xjcut is not None:
                    ax3.axvline(x,
                                c=col, ls='--')
        hand = [mlines.Line2D([], [], color=lcol[ii % ncol], ls='--')
                for ii in range(nions)]
        axs2.legend(hand, lions,
                    bbox_to_anchor=(1., 1.02), loc='upper left')

    # Plot 1d spectra and associated phi windows
    for ii in range(nspect):
        axs2.plot(
            lambfit, spect1d[ii],
            c=lcolspect[ii % ncolspect],
            ls='-',
            )
        ax2.axhline(
            phiminmax[ii, 0],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax2.axhline(
            phiminmax[ii, 1],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax3.axhline(
            phiminmax[ii, 0],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )
        ax3.axhline(
            phiminmax[ii, 1],
            c=lcolspect[ii % ncolspect],
            ls='-', lw=1.,
            )

    ax2.set_xlim(lambmin, lambmax)
    ax2.set_ylim(phifit.min(), phifit.max())
    plt.setp(ax1.get_xticklabels(), visible=False)
    plt.setp(ax1.get_yticklabels(), visible=False)
    plt.setp(ax2.get_xticklabels(), visible=False)
    plt.setp(ax3.get_yticklabels(), visible=False)

    if tit is not False:
        fig.suptitle(tit, size=14, weight='bold')
    if wintit is not False:
        fig.canvas.manager.set_window_title(wintit)
    return [ax0, ax1]


# #################################################################
# #################################################################
#                   fit1d_dinput plot
# #################################################################
# #################################################################


def _domain2str(domain, key):
    return [
        ss.tolist() if isinstance(ss, np.ndarray)
        else ss
        for ss in domain[key]['spec']
    ]


def _check_phi_lim(phi=None, phi_lim=None, islamb=None):

    Dphi = (phi.max()-phi.min()) / 2.
    dphi = np.mean(np.diff(np.unique(phi)))
    if phi_lim is None:
        if islamb is True:
            phi_lim = phi.mean() + Dphi*np.r_[-0.56, -0.04, 0.59]
        else:
            phi_lim = phi.mean() + Dphi*np.r_[0., -0.25, -0.5]

    # check for phi values
    c0 = (
        np.isscalar(phi_lim)
        or (
            hasattr(phi_lim, '__iter__')
            and len(phi_lim) != 2
            and all([np.isscalar(pp) for pp in phi_lim])
        )
    )
    if c0:
        if np.isscalar(phi_lim):
            phi_lim = [phi_lim]
        phi_lim = dphi * np.r_[-0.6, 0.6][None, :] + np.r_[phi_lim][:, None]

    # check
    c0 = (
        hasattr(phi_lim, '__iter__')
        and (
            (
                len(phi_lim) == 2
                and all([np.isscalar(pp) for pp in phi_lim])
                and phi_lim[0] < phi_lim[1]
            )
            or (
                all([
                    hasattr(pp, '__iter__')
                    and len(pp) == 2
                    and all([np.isscalar(ppi) for ppi in pp])
                    and pp[0] < pp[1]
                    for pp in phi_lim
                ])
            )
        )
    )
    if not c0:
        msg = (
            "Arg phi_lim must be a an iterable of iterables of len() == 2\n"
            f"Provided: {phi_lim}"
        )
        raise Exception(msg)

    if all([np.isscalar(pp) for pp in phi_lim]):
        phi_lim = [phi_lim]
    phi_lim = np.asarray(phi_lim)

    return phi_lim


def plot_dinput2d(
    # input data
    dinput=None,
    indspect=None,
    phi_lim=None,
    phi_name=None,
    lamb_lim=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
    wintit=None,
    cmap=None,
    vmin=None,
    vmax=None,
    cmap_indok=None,
):

    # Check inputs
    # ------------

    # phi
    phi_lim = _check_phi_lim(phi=dinput['dprepare']['phi'], phi_lim=phi_lim)
    if phi_name is None:
        phi_name = r'$\phi$'

    # lamb
    lamb_lim = _check_phi_lim(
        phi=dinput['dprepare']['lamb'],
        phi_lim=lamb_lim,
        islamb=True,
    )

    if fs is None:
        fs = (13, 6)
    if wintit is None:
        wintit = _WINTIT
    if dmargin is None:
        dmargin = {
            'left': 0.08, 'right': 0.97,
            'bottom': 0.07, 'top': 0.95,
            'wspace': 1., 'hspace': 0.4,
        }

    if vmin is None:
        vmin = 0.
    if vmax is None:
        vmax = np.nanmax(dinput['dprepare']['data'])

    if cmap_indok is None:
        cmap_indok = plt.cm.Accent_r

    if indspect is None:
        indspect = dinput['valid']['indt'].nonzero()[0][0]
    indspect = np.atleast_1d(indspect).ravel()
    if indspect.dtype == 'bool':
        indspect = np.nonzero(indspect)[0]
    nspect = indspect.size
    nspecttot = dinput['dprepare']['data'].shape[0]

    # Extract (better redeability)
    dprepare = dinput['dprepare']

    # Extract data
    lamb = dprepare['lamb']
    phi = dprepare['phi']
    data = dprepare['data'][indspect, ...]

    # indok
    indok = dprepare['indok'][indspect, ...]
    # add valid
    nbs = dinput['nbs']

    # Extent
    extent = (lamb.min(), lamb.max(), phi.min(), phi.max())

    # Extent
    if dprepare['binning'] is False:
        nxi, nxj = data.shape[1:]
        extent = (0, nxi, 0, nxj)
        x0_lab = 'xi'
        x1_lab = 'xj'
        aspect = 'equal'
    else:
        nlamb, nphi = data.shape[1:]
        extent = (lamb.min(), lamb.max(), phi.min(), phi.max())
        x0_lab = r'$\lambda$'
        x1_lab = phi_name
        aspect = 'auto'

    # indok lab
    indok_lab = [
        dprepare['dindok'].get(ii, ii) for ii in range(-cmap_indok.N + 1, 1)
    ]

    # Plot
    # ------------

    ldax = []
    lcol = ['r', 'b', 'g', 'y', 'm', 'c']
    for ii, ispect in enumerate(indspect):
        if tit is None:
            titi = f"spect {ispect} ({ii+1}/{nspect} out of {nspecttot})"
        else:
            titi = tit

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(3, 19, **dmargin)

        # axes for pixel validity
        ax0 = fig.add_subplot(gs[:2, :5])
        cax0 = fig.add_subplot(gs[:2, 5])
        ax0.set_ylabel(x1_lab, size=12, fontweight='bold')
        ax0.set_xlabel(x0_lab, size=12, fontweight='bold')
        titax = f"indt ok: {dinput['valid']['indt'][ispect]}"
        ax0.set_title(titax, size=12, fontweight='bold')

        # axes for image
        ax1 = fig.add_subplot(gs[:2, 9:14], sharex=ax0, sharey=ax0)
        cax1 = fig.add_subplot(gs[:2, 14])

        # axes for colorbar
        ax2 = fig.add_subplot(gs[:2, 8], sharey=ax0)
        ax2.set_title('knots', size=12, fontweight='bold')

        # axes for spectra
        ax3 = fig.add_subplot(gs[2, 9:14], sharex=ax0)
        ax3.set_ylabel('data (a.u.)', size=12, fontweight='bold')
        ax3.set_xlabel(x0_lab, size=12, fontweight='bold')
        ax3.set_title('data', size=12, fontweight='bold')

        # axes for vertical slices
        ax4 = fig.add_subplot(gs[:2, 16:], sharey=ax0)
        ax4.set_ylabel(x1_lab, size=12, fontweight='bold')
        ax4.set_xlabel('data (a.u.)', size=12, fontweight='bold')
        ax4.set_title('vertical slices', size=12, fontweight='bold')

        # dict
        dax = {
            'img_indok': {'ax': ax0},
            'img_original_data': {'ax': ax1},
            'bsplines': {'ax': ax2},
            'spect': {'ax': ax3},
            'vert': {'ax': ax4},
        }

        # plot images
        kax = 'img_indok'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            im = ax.imshow(
                indok[ii, ...],
                extent=extent,
                cmap=cmap_indok,
                vmin=-cmap_indok.N + 0.5,
                vmax=0.5,
                interpolation='nearest',
                origin='lower',
                aspect=aspect,
            )
            fig.colorbar(
                im,
                ax=ax,
                cax=cax0,
                ticks=range(-cmap_indok.N+1, 1),
                orientation='vertical',
            )
            cax0.set_yticklabels(indok_lab)
            for jj in range(len(dinput['valid']['ldphi'][ispect])):
                ax.axhline(
                    dinput['valid']['ldphi'][ispect][jj][0],
                    c='k', lw=2., ls='-',
                )
                ax.axhline(
                    dinput['valid']['ldphi'][ispect][jj][1],
                    c='k', lw=2., ls='-',
                )

        kax = 'img_original_data'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            im = ax.imshow(
                data[ii, ...],
                extent=extent,
                cmap=cmap,
                vmin=vmin,
                vmax=vmax,
                interpolation='nearest',
                origin='lower',
                aspect=aspect,
            )

            fig.colorbar(im, ax=ax, cax=cax1, orientation='vertical')
            for jj in range(phi_lim.shape[0]):
                ax.axhline(phi_lim[jj, 0], c=lcol[jj], lw=1., ls='-')
                ax.axhline(phi_lim[jj, 1], c=lcol[jj], lw=1., ls='-')
            for jj in range(lamb_lim.shape[0]):
                ax.axvline(lamb_lim[jj, 0], c=lcol[jj], lw=1., ls='-')
                ax.axvline(lamb_lim[jj, 1], c=lcol[jj], lw=1., ls='-')

            for jj in range(len(dinput['valid']['ldphi'][ispect])):
                ax.axhline(
                    dinput['valid']['ldphi'][ispect][jj][0],
                    c='k', lw=2., ls='-',
                )
                ax.axhline(
                    dinput['valid']['ldphi'][ispect][jj][1],
                    c='k', lw=2., ls='-',
                )

        kax = 'bsplines'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            for jj, jk in enumerate(np.unique(dinput['knots_mult'])):
                ax.axhline(jk, c='k', ls='-', lw=1.)

        # spectra
        kax = 'spect'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            for jj in range(phi_lim.shape[0]):
                indphi = (phi >= phi_lim[jj, 0]) & (phi < phi_lim[jj, 1])
                spect_lamb = lamb[indphi]
                indlamb = np.argsort(spect_lamb)
                spect_lamb = spect_lamb[indlamb]
                spect_data = data[ii, indphi][indlamb]
                ax.plot(
                    spect_lamb,
                    spect_data,
                    marker='.',
                    ms=3,
                    color=lcol[jj],
                    ls='None',
                )
            ax.axhline(0, c='k', lw=1., ls='-')

            # add lines
            for jj, jk in enumerate(dinput['lines']):
                ax.axvline(jk, c='k', ls='--', lw=1.)
                ax.text(
                    jk, 1, dinput['symb'][jj],
                    transform=ax.get_xaxis_transform(),
                    verticalalignment='bottom',
                    horizontalalignment='center',
                )

        # vertical slices
        kax = 'vert'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            for jj in range(lamb_lim.shape[0]):
                indlamb = (lamb >= lamb_lim[jj, 0]) & (lamb < lamb_lim[jj, 1])
                vert_phi = phi[indlamb]
                indphi = np.argsort(vert_phi)
                vert_phi = vert_phi[indphi]
                vert_data = data[ii, indlamb][indphi]
                ax.plot(
                    vert_data,
                    vert_phi,
                    marker='.',
                    ms=3,
                    color=lcol[jj],
                    ls='None',
                )
            ax.axvline(0, c='k', lw=1., ls='-')

        # text
        binstr = dprepare['binning']
        if binstr is False:
            binstr = (False, False)
        else:
            binstr = (binstr['lamb']['nbins'], binstr['phi']['nbins'])
        focusstr = dinput['valid']['focus']
        if focusstr is not False:
            focusstr = focusstr.tolist()
        msg = (
            f"symmetry: {dinput['symmetry'] is True}\n\n"
            "domain:\n"
            f"    'lamb' = {_domain2str(dprepare['domain'], 'lamb')}\n"
            f"    'phi' = {_domain2str(dprepare['domain'], 'phi')}\n\n"
            "Binning:\n"
            f"    'lamb': {binstr[0]}\n"
            f"    'phi': {binstr[1]}\n\n"
            "S/N:\n"
            f"    nsigma: {dinput['valid']['valid_nsigma']}\n"
            f"    fraction: {dinput['valid']['valid_fraction']}\n"
            f"    focus: {focusstr}\n"
            f"    nbs: {dinput['valid']['indbs'][ispect, :].sum()} / {nbs}\n"
        )
        fig.text(
            0.04,
            0.29,
            msg,
            color='k',
            fontsize=8,
            verticalalignment='top',
        )

        ldax.append(dax)

    return ldax


# #################################################################
# #################################################################
#                   fit1d plot
# #################################################################
# #################################################################


def plot_fit1d(
    # input data
    dfit1d=None,
    dextract=None,
    # options
    annotate=None,
    showonly=None,
    indspect=None,
    xlim=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
    wintit=None,
):

    # Check inputs
    # ------------
    if annotate is None:
        annotate = True
    if annotate is True:
        annotate = dfit1d['dinput']['keys']
    if isinstance(annotate, str):
        annotate = [annotate]
    if xlim is None:
        xlim = False
    if fs is None:
        fs = (15, 8)
    if wintit is None:
        wintit = _WINTIT
    if dmargin is None:
        dmargin = {'left': 0.05, 'right': 0.85,
                   'bottom': 0.07, 'top': 0.90,
                   'wspace': 0.2, 'hspace': 0.3}

    # Extract (better redeability)
    dprepare = dfit1d['dinput']['dprepare']
    dinput = dfit1d['dinput']
    d3 = dextract['d3']

    lamb = dprepare['lamb']
    data = dprepare['data']

    # Index of spectra to plot
    if not np.any(dinput['valid']['indt']):
        msg = "The provided fit1d result has no valid time step!"
        raise Exception(msg)

    if indspect is None:
        indspect = dinput['valid']['indt'].nonzero()[0][0]
    indspect = np.atleast_1d(indspect).ravel()
    if indspect.dtype == 'bool':
        indspect = np.nonzero(indspect)[0]

    # pre-compute
    # ------------
    if dinput['same_spectrum'] is True:
        nlines = int(dinput['nlines'] / dinput['same_spectrum_nspect'])
        dinput['lines'] = dinput['lines'][:nlines]
        dinput['ion'] = dinput['ion'][:nlines]
        dinput['symb'] = dinput['symb'][:nlines]
        nwidth = int(dinput['width']['keys'].size
                     / dinput['same_spectrum_nspect'])
        dinput['width']['keys'] = dinput['width']['keys'][:nwidth]
        dinput['width']['ind'] = dinput['width']['ind'][:, :nlines]
        dinput['width']['ind'] = dinput['width']['ind'][:nwidth, :]
        dinput['shift']['ind'] = dinput['shift']['ind'][:, :nlines]

    nlines = dinput['lines'].size
    ions_u = sorted(set(dinput['ion'].tolist()))
    nions = len(ions_u)

    indwidth = np.argmax(dinput['width']['ind'], axis=0)
    indshift = np.argmax(dinput['shift']['ind'], axis=0)

    x = dinput['lines'][None, :] + d3['shift']['lines']['values']

    lcol = ['k', 'r', 'b', 'g', 'm', 'c']
    ncol = len(lcol)
    Ti = 'Ti' in d3.keys()
    vi = 'vi' in d3.keys()
    ratio = 'ratio' in d3.keys()

    if Ti:
        lfcol = ['y', 'g', 'c', 'm']
    else:
        lfcol = [None]
    nfcol = len(lfcol)
    if vi:
        lhatch = [None, '/', '\\', '|', '-', '+', 'x', '//']
    else:
        lhatch = [None]
    nhatch = len(lhatch)
    nspect = indspect.size

    # Plot
    # ------------

    for ii in range(nspect):
        ispect = indspect[ii]
        if tit is None:
            titi = ("spect {}\n".format(ispect+1)
                    + "({}/{})".format(ispect, dprepare['data'].shape[0]))
        else:
            titi = tit

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(4, 1, **dmargin)
        ax0 = fig.add_subplot(gs[:-1, 0])
        ax1 = fig.add_subplot(gs[-1, 0], sharex=ax0)

        ax0.set_ylabel(r'data (a.u.)')
        ax1.set_ylabel(r'error (a.u.)')
        ax1.set_xlabel(r'$\lambda$ (m)')

        ax0.plot(
            lamb[dprepare['indok_bool'][ispect, :]],
            data[ispect, dprepare['indok_bool'][ispect, :]],
            marker='.', c='k', ls='None', ms=8,
        )
        ax0.plot(
            lamb[~dprepare['indok_bool'][ispect, :]],
            data[ispect, ~dprepare['indok_bool'][ispect, :]],
            marker='x', c='k', ls='None', ms=4,
        )
        if showonly is not True:
            if dextract['sol_detail'] is not False:
                ax0.plot(
                    dprepare['lamb'], dextract['sol_detail'][ispect, :, 0],
                    ls='-', c='k',
                )
            ax0.set_prop_cycle(None)
            if dextract['sol_detail'] is not False:
                if Ti or vi:
                    for jj in range(nlines):
                        col = lfcol[indwidth[jj] % nfcol]
                        hatch = lhatch[indshift[jj] % nhatch]
                        ax0.fill_between(
                            dprepare['lamb'],
                            dextract['sol_detail'][ispect, :, 1+jj],
                            alpha=0.3, color=col, hatch=hatch,
                        )
                else:
                    ax0.plot(
                        dprepare['lamb'],
                        dextract['sol_detail'][ispect, :, 1:].T,
                    )
            if dextract['sol_total'] is not False:
                ax0.plot(
                    dprepare['lamb'],
                    dextract['sol_total'][ispect, :],
                    c='k', lw=2.,
                )
                ax1.plot(
                    dprepare['lamb'],
                    dextract['sol_total'][ispect, :] - dprepare['data'][ispect, :],
                    c='k', lw=2.,
                )
                ax1.axhline(0, c='k', ls='--')

        # Annotate lines
        if annotate is not False:
            for jj, k0 in enumerate(ions_u):
                col = lcol[jj % ncol]
                ind = (dinput['ion'] == k0).nonzero()[0]
                for nn in ind:
                    if dinput['keys'][nn] not in annotate:
                        continue
                    lab = dinput['symb'][nn]
                    ax0.axvline(x[ispect, nn], c=col, ls='--')
                    if 'x' in d3['amp'].keys():
                        val = d3['amp']['lines']['values'][ispect, nn]
                        lab += '\n{:4.2e}'.format(val)
                    if 'x' in d3['shift'].keys():
                        val = d3['shift']['lines']['values'][ispect, nn]*1.e10
                        lab += '\n({:+4.2e} A)'.format(val)
                    ax0.annotate(
                        lab,
                        xy=(x[ispect, nn], 1.01), xytext=None,
                        xycoords=('data', 'axes fraction'),
                        color=col, arrowprops=None,
                        horizontalalignment='center',
                        verticalalignment='bottom',
                    )

            # Ion legend
            hand = [
                mlines.Line2D(
                    [], [],
                    color=lcol[jj % ncol], ls='--', label=ions_u[jj]
                )
                for jj in range(nions)
            ]
            legi = ax0.legend(
                handles=hand,
                title='ions',
                bbox_to_anchor=(1.01, 1.),
                loc='upper left',
            )
            ax0.add_artist(legi)

        # Ti legend
        if Ti:
            hand = [
                mpatches.Patch(color=lfcol[jj % nfcol])
                for jj in range(dinput['width']['ind'].shape[0])
            ]
            lleg = [
                str(dinput['width']['keys'][jj])
                + '{:4.2f}'.format(
                    d3['Ti']['lines']['values'][ispect, jj]*1.e-3
                )
                for jj in range(dinput['width']['ind'].shape[0])
            ]
            legT = ax0.legend(
                handles=hand, labels=lleg,
                title='Ti (keV)',
                bbox_to_anchor=(1.01, 1.),    # 0.8
                loc='upper left',
            )
            ax0.add_artist(legT)

        # vi legend
        if vi:
            hand = [
                mpatches.Patch(
                    facecolor='w', edgecolor='k',
                    hatch=lhatch[jj % nhatch],
                )
                for jj in range(dinput['shift']['ind'].shape[0])
            ]
            lleg = [
                str(dinput['shift']['keys'][jj])
                + '{:4.2f}'.format(
                    d3['vi']['x']['values'][ispect, jj]*1.e-3
                )
                for jj in range(dinput['shift']['ind'].shape[0])
            ]
            legv = ax0.legend(
                handles=hand, labels=lleg,
                title='vi (km/s)',
                bbox_to_anchor=(1.01, 0.75),    # 0.5
                loc='upper left',
            )
            ax0.add_artist(legv)

        # Ratios legend
        if ratio:
            nratio = d3['ratio']['lines']['values'].shape[1]
            hand = [mlines.Line2D([], [], c='k', ls='None')]*nratio
            lleg = ['{} =  {:4.2e}'.format(
                d3['ratio']['lines']['lab'][jj],
                d3['ratio']['lines']['values'][ispect, jj])
                    for jj in range(nratio)]
            legr = ax0.legend(
                handles=hand,
                labels=lleg,
                title='line ratio',
                bbox_to_anchor=(1.01, 0.30),    # 0.21
                loc='lower left',
            )
            ax0.add_artist(legr)

        # bck legend
        if True:
            hand = [mlines.Line2D([], [], c='k', ls='None')]*2
            lleg = [
                f"amp = {d3['bck_amp']['x']['values'][ispect, 0]:4.2e}",
                f"rate = {d3['bck_rate']['x']['values'][ispect, 0]:4.2e}"
            ]
            legr = ax0.legend(
                handles=hand,
                labels=lleg,
                title='background',
                bbox_to_anchor=(1.01, 0.10),    # 0.05
                loc='lower left',
            )
            ax0.add_artist(legr)

        # double legend
        if dinput['double'] is not False:
            hand = [mlines.Line2D([], [], c='k', ls='None')]*2
            lleg = [
                f"ratio = {d3['dratio']['x']['values'][ispect, 0]:4.2f}",
                'shift ' + r'$\approx$'
                + f" {d3['dshift']['x']['values'][ispect, 0]:4.2e}"
            ]
            legr = ax0.legend(
                handles=hand,
                labels=lleg,
                title='double',
                bbox_to_anchor=(1.01, -0.1),
                loc='lower left',
            )

        ax0.set_xlim(dinput['dprepare']['domain']['lamb']['minmax'])

        if titi is not False:
            fig.suptitle(titi, size=14, weight='bold')
        if wintit is not False:
            fig.canvas.manager.set_window_title(wintit)
    if xlim is not False:
        ax0.set_xlim(xlim)

    return {'data': ax0, 'error': ax1}


def plot_fit2d(
    # input data
    dfit2d=None,
    dextract=None,
    # options
    annotate=None,
    showonly=None,
    indspect=None,
    xlim=None,
    phi_lim=None,
    phi_name=None,
    # figure
    fs=None,
    dmargin=None,
    tit=None,
    wintit=None,
    cmap=None,
    vmin=None,
    vmax=None,
    vmin_err=None,
    vmax_err=None,
):

    # Check inputs
    # ------------
    if annotate is None:
        annotate = True
    if annotate is True:
        annotate = dfit2d['dinput']['keys']
    if isinstance(annotate, str):
        annotate = [annotate]
    if xlim is None:
        xlim = False
    phi_lim = _check_phi_lim(
        phi=dfit2d['dinput']['dprepare']['phi'], phi_lim=phi_lim,
    )
    if phi_name is None:
        phi_name = r'$phi$'

    if fs is None:
        fs = (16, 9)
    if wintit is None:
        wintit = _WINTIT
    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.07, 'top': 0.90,
            'wspace': 0.2, 'hspace': 0.5,
        }

    if vmin is None:
        vmin = 0.
    if vmax is None:
        vmax = np.nanmax(dfit2d['dinput']['dprepare']['data'])

    # Extract (better redeability)
    dprepare = dfit2d['dinput']['dprepare']
    dinput = dfit2d['dinput']
    d3 = dextract['d3']

    # Index of spectra to plot
    if not np.any(dinput['valid']['indt']):
        msg = "The provided fit1d result has no valid time step!"
        raise Exception(msg)

    if indspect is None:
        indspect = dinput['valid']['indt'].nonzero()[0][0]
    indspect = np.atleast_1d(indspect).ravel()
    if indspect.dtype == 'bool':
        indspect = np.nonzero(indspect)[0]

    nlines = dinput['lines'].size
    ions_u = sorted(set(dinput['ion'].tolist()))
    nions = len(ions_u)

    indwidth = np.argmax(dinput['width']['ind'], axis=0)
    indshift = np.argmax(dinput['shift']['ind'], axis=0)

    x = dinput['lines'][None, :] + d3['shift']['lines']['values']

    lcol = ['k', 'r', 'b', 'g', 'm', 'c']
    ncol = len(lcol)
    Ti = 'Ti' in d3.keys()
    vi = 'vi' in d3.keys()
    ratio = 'ratio' in d3.keys()

    if Ti:
        lfcol = ['y', 'g', 'c', 'm']
    else:
        lfcol = [None]
    nfcol = len(lfcol)
    if vi:
        lhatch = [None, '/', '\\', '|', '-', '+', 'x', '//']
    else:
        lhatch = [None]
    nhatch = len(lhatch)
    nspect = indspect.size
    nspecttot = dprepare['data'].shape[0]

    # Extract data
    lamb = dprepare['lamb']
    phi = dprepare['phi']
    data = dprepare['data'][indspect, ...]
    sol_tot = dextract['sol_tot'][indspect, ...]

    # set to nan if not indok
    for ii, ispect in enumerate(indspect):
        data[ii, ~dprepare['indok_bool'][ispect, ...]] = np.nan
        sol_tot[ii, ~dprepare['indok_bool'][ispect, ...]] = np.nan

    if np.any(dextract['indphi'][indspect, :]):
        dphi = np.tile(dextract['phi_prof'], (nspect, 1))
        dphi = np.atleast_2d(dphi[dextract['indphi'][indspect, :]])
        dphi = np.array([np.min(dphi, axis=1), np.max(dphi, axis=1)]).T
    else:
        dphi = None

    # Error if relevant
    err = sol_tot - data
    if vmin_err is None or vmax_err is None:
        v_err = np.nanmax(np.abs(err))
        if vmin_err is None:
            vmin_err = -v_err
        if vmax_err is None:
            vmax_err = v_err

    # Extent
    extent = (lamb.min(), lamb.max(), phi.min(), phi.max())

    # 1d slice
    lspect_lamb = []
    lspect_data = []
    lspect_fit = []
    lspect_err = []
    for jj in range(phi_lim.shape[0]):
        indphi = (phi >= phi_lim[jj, 0]) & (phi < phi_lim[jj, 1])
        spect_lamb = lamb[indphi]
        indlamb = np.argsort(spect_lamb)
        lspect_lamb.append(spect_lamb[indlamb])
        lspect_data.append(data[:, indphi][:, indlamb])
        lspect_fit.append(dextract['func_tot'](
            lamb=lspect_lamb[jj],
            phi=phi[indphi][indlamb],
        )[indspect, ...])
        lspect_err.append(lspect_fit[jj] - lspect_data[jj])

    # Plot
    # ------------

    ldax = []
    lcol_spect = ['r', 'b', 'g', 'y', 'm', 'c']
    for ii, ispect in enumerate(indspect):
        if tit is None:
            titi = f"spect {ispect} ({ii+1}/{nspect} out of {nspecttot})"
        else:
            titi = tit

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(7, 9, **dmargin)
        ax0 = fig.add_subplot(gs[:4, :2])
        ax1 = fig.add_subplot(gs[:4, 2:4], sharex=ax0, sharey=ax0)
        ax2 = fig.add_subplot(gs[:4, 4:6], sharex=ax0, sharey=ax0)
        ax3 = fig.add_subplot(gs[:4, 6], sharey=ax0)
        ax4 = fig.add_subplot(gs[:4, 7], sharey=ax0)
        ax5 = fig.add_subplot(gs[:4, 8], sharey=ax0)
        ax6 = fig.add_subplot(gs[4:6, :6], sharex=ax0)
        ax7 = fig.add_subplot(gs[6, :6], sharex=ax0)

        ax0.set_ylabel(phi_name)
        ax0.set_xlabel(r'$\lambda$ (m)')

        ax3.set_xlabel('line ratios')
        ax4.set_xlabel(r'$T_{i}$ (keV)')
        ax5.set_xlabel(r'$v_{rot}$ (km/s)')

        ax6.set_ylabel('data (a.u.)')
        ax7.set_ylabel('err (a.u.)')
        ax7.set_xlabel(r'$\lambda$ (m)')

        # remove yticklabels
        # ax1.get_yaxis().set_visible(False)
        plt.setp(ax1.get_yticklabels(), visible=False)
        plt.setp(ax2.get_yticklabels(), visible=False)

        plt.setp(ax4.get_yticklabels(), visible=False)
        plt.setp(ax5.get_yticklabels(), visible=False)

        plt.setp(ax6.get_xticklabels(), visible=False)

        dax = {
            'img_data': {'ax': ax0},
            'img_fit': {'ax': ax1},
            'img_err': {'ax': ax2},
            'prof_Te': {'ax': ax3},
            'prof_Ti': {'ax': ax4},
            'prof_vi': {'ax': ax5},
            'spect': {'ax': ax6},
            'spect_err': {'ax': ax7},
        }

        # plot images
        kax = 'img_data'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            im = ax.imshow(
                data[ii, ...],
                extent=extent,
                interpolation='nearest',
                origin='lower',
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
                aspect='auto',
            )
            fig.colorbar(im, ax=ax, orientation='vertical')
            # for jj in range(phi_lim.shape[0]):
            # ax.axhline(phi_lim[jj, 0], c=lcol_spect[jj], lw=1., ls='-')
            # ax.axhline(phi_lim[jj, 1], c=lcol_spect[jj], lw=1., ls='-')

        kax = 'img_fit'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            ax.imshow(
                sol_tot[ii, ...],
                extent=extent,
                interpolation='nearest',
                origin='lower',
                vmin=vmin,
                vmax=vmax,
                cmap=cmap,
                aspect='auto',
            )
            for jj in range(phi_lim.shape[0]):
                ax.axhline(phi_lim[jj, 0], c=lcol_spect[jj], lw=1., ls='-')
                ax.axhline(phi_lim[jj, 1], c=lcol_spect[jj], lw=1., ls='-')

        kax = 'img_err'
        if dax.get(kax) is not None:
            ax = dax[kax]['ax']
            im = ax.imshow(
                err[ii, ...],
                extent=extent,
                interpolation='nearest',
                origin='lower',
                vmin=vmin_err,
                vmax=vmax_err,
                cmap=plt.cm.seismic,
                aspect='auto',
            )
            plt.colorbar(im, ax=ax, orientation='vertical')

        kax, k1 = 'prof_Te', 'ratio'
        if dax.get(kax) is not None and 'ratio' in d3.keys():
            ax = dax[kax]['ax']
            for jj in range(d3['ratio']['requested'].shape[-1]):
                ax.plot(
                    d3['ratio']['lines']['values'][ispect, :, jj],
                    dextract['phi_prof'],
                    ls='-',
                    lw=1,
                    label=d3['ratio']['lines']['lab'][jj],
                )
            ax.axvline(0, c='k', ls='--', lw=1.)
            if dphi is not None:
                ax.axhline(dphi[ii, 0], c='k', ls='-', lw=1.)
                ax.axhline(dphi[ii, 1], c='k', ls='-', lw=1.)

        kax, k1 = 'prof_Ti', 'Ti'
        if dax.get(kax) is not None and k1 in d3.keys():
            ax = dax[kax]['ax']
            for jj in range(d3[k1]['lines']['values'].shape[-1]):
                ax.plot(
                    d3[k1]['lines']['values'][ispect, :, jj]*1e-3,
                    dextract['phi_prof'],
                    ls='-',
                    lw=1,
                    label=d3[k1]['lines']['keys'][jj],
                )
            ax.axvline(0, c='k', ls='--', lw=1.)
            if dphi is not None:
                ax.axhline(dphi[ii, 0], c='k', ls='-', lw=1.)
                ax.axhline(dphi[ii, 1], c='k', ls='-', lw=1.)

        kax, k1 = 'prof_vi', 'vi'
        if dax.get(kax) is not None and k1 in d3.keys():
            ax = dax[kax]['ax']
            for jj in range(d3[k1]['x']['values'].shape[-1]):
                ax.plot(
                    d3[k1]['x']['values'][ispect, :, jj]*1.e-3,
                    dextract['phi_prof'],
                    ls='-',
                    lw=1,
                    label=d3[k1]['x']['keys'][jj],
                )
            ax.axvline(0, c='k', ls='--', lw=1.)
            if dphi is not None:
                ax.axhline(dphi[ii, 0], c='k', ls='-', lw=1.)
                ax.axhline(dphi[ii, 1], c='k', ls='-', lw=1.)
            # adjust
            ax.set_ylim(phi.min(), phi.max())
            ax.legend()

        kax = 'spect'
        if dax.get(kax) is not None and k1 in d3.keys():
            ax = dax[kax]['ax']
            for jj in range(phi_lim.shape[0]):
                ax.plot(
                    lspect_lamb[jj],
                    lspect_data[jj][ii, ...],
                    c=lcol_spect[jj],
                    ls='None',
                    marker='.',
                )
                ax.plot(
                    lspect_lamb[jj],
                    lspect_fit[jj][ii, ...],
                    ls='-',
                    c=lcol_spect[jj],
                    lw=1.,
                )

        kax = 'spect_err'
        if dax.get(kax) is not None and k1 in d3.keys():
            ax = dax[kax]['ax']
            for jj in range(phi_lim.shape[0]):
                ax.plot(
                    lspect_lamb[jj], lspect_err[jj][ii, ...],
                    marker='.', c=lcol_spect[jj],
                )
            ax.axhline(0, ls='--', c='k', lw=1.)

        ldax.append(dax)

    return ldax


# DEPRECATED - keeping for back-up
def CrystalBragg_plot_data_fit2d(xi, xj, data, lamb, phi, indok=None,
                                 dfit2d=None,
                                 dax=None, indspect=None,
                                 spect1d=None, fit1d=None,
                                 lambfit=None, phiminmax=None,
                                 cmap=None, vmin=None, vmax=None,
                                 plotmode=None, fs=None, dmargin=None,
                                 angunits='deg', tit=None, wintit=None):

    # Check inputs
    # ------------

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if cmap is None:
            cmap = plt.cm.viridis
        if dmargin is None:
            dmargin = {'left': 0.05, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.95,
                       'wspace': None, 'hspace': 0.8}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
    if dfit2d['dinput']['symmetry'] is True:
        symaxis = dfit2d['dinput']['symmetry_axis'][indspect]
    if angunits is None:
        angunits = 'deg'
    assert angunits in ['deg', 'rad']

    phiflat = dfit2d['phi']
    pts_phi = dfit2d['pts_phi']
    ylim = np.r_[dfit2d['dinput']['phiminmax']]
    if angunits == 'deg':
        phiflat = phiflat*180./np.pi
        phi = phi*180./np.pi
        pts_phi = pts_phi*180./np.pi
        ylim = ylim*180./np.pi
        phiminmax = phiminmax*180./np.pi
        if dfit2d['dinput']['symmetry'] is True:
            symaxis = symaxis*180./np.pi
    if plotmode == 'transform':
        xlab = r'$\lambda$ (m)'
        ylab = r'$\phi$ ({})'.format(angunits)
    else:
        xlab = r'$x_i$' + r' (m)'
        ylab = r'$x_j$' + r' ({})'.format(angunits)
        spect1d = None

    # pre-compute
    # ------------

    if vmin is None:
        vmin = 0.
    if vmax is None:
        vmax = max(np.nanmax(dfit2d['data'][indspect, :]),
                   np.nanmax(dfit2d['sol_total'][indspect, :]))
    err = dfit2d['sol_total'][indspect, :] - dfit2d['data'][indspect, :]
    errm = vmax/10.

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        naxh = (3*3
                + 2*(dfit2d['dinput']['Ti'] + dfit2d['dinput']['vi']
                     + (dfit2d['ratio'] not in [None, False])))
        naxv = 1
        gs = gridspec.GridSpec((3+naxv)*3, naxh, **dmargin)
        if plotmode == 'transform':
            ax0 = fig.add_subplot(gs[:9, :3])
            ax0.set_xlim(dfit2d['dinput']['lambminmax'])
            ax0.set_ylim(ylim)
        else:
            ax0 = fig.add_subplot(gs[:9, :3], aspect='equal')
        ax0c = fig.add_subplot(gs[10, :3])
        ax1 = fig.add_subplot(
            gs[:9, 3:6],
            sharex=ax0, sharey=ax0,
        )
        ax2 = fig.add_subplot(
            gs[:9, 6:9],
            sharex=ax0, sharey=ax0,
        )
        ax1c = fig.add_subplot(gs[10, 3:6])

        ax0.set_title('Residu')
        ax0.set_xlabel(xlab)
        ax0.set_ylabel(ylab)
        ax1.set_title('Original data')
        ax1.set_xlabel(xlab)
        ax2.set_title('2d fit')
        dax = {'err': ax0, 'err_colorbar': ax0c,
               'data': ax1, 'data_colorbar': ax1c,
               'fit': ax2}
        if tit is not False:
            fig.suptitle(tit)
        if wintit is not False:
            fig.canvas.manager.set_window_title(wintit)

        if naxv == 1:
            dax['fit1d'] = fig.add_subplot(gs[9:11, 6:9], sharex=ax2)
            dax['fit1d'].set_title('1d sliced spectrum vs fit')
            dax['fit1d'].set_ylabel(r'data')
            dax['err1d'] = fig.add_subplot(gs[11, 6:9], sharex=ax2)
            dax['err1d'].set_ylabel(r'err')
            dax['err1d'].set_xlabel(xlab)

        nn = 1
        if dfit2d['dinput']['Ti'] is True:
            dax['Ti'] = fig.add_subplot(gs[:9, 9+2*(nn-1):9+2*nn], sharey=ax1)
            dax['Ti'].set_title(r'Width')
            dax['Ti'].set_xlabel(r'$\hat{T_i}$' + r' (keV)')
            nn += 1
        if dfit2d['dinput']['vi'] is True:
            dax['vi'] = fig.add_subplot(gs[:9, 9+2*(nn-1):9+2*nn], sharey=ax1)
            dax['vi'].set_title(r'Shift')
            dax['vi'].set_xlabel(r'$\hat{v_i}$' + r' (km/s)')
            dax['vi'].axvline(0, ls='-', lw=1., c='k')
            nn += 1
        if dfit2d['ratio'] not in [None, False]:
            dax['ratio'] = fig.add_subplot(gs[:9, 9+2*(nn-1):], sharey=ax1)
            dax['ratio'].set_title(r'Intensity Ratio')
            dax['ratio'].set_xlabel(r'ratio (a.u)')

    # Plot main images
    # ------------
    if plotmode == 'transform':
        if dax.get('err') is not None:
            errax = dax['err'].scatter(dfit2d['lamb'],
                                       phiflat,
                                       c=err,
                                       s=6, marker='s', edgecolors='None',
                                       vmin=-errm, vmax=errm,
                                       cmap=plt.cm.seismic)
        if dax.get('data') is not None:
            dataax = dax['data'].scatter(dfit2d['lamb'],
                                         phiflat,
                                         c=dfit2d['data'][indspect, :],
                                         s=6, marker='s', edgecolors='None',
                                         vmin=vmin, vmax=vmax, cmap=cmap)
        if dax.get('fit') is not None:
            dax['fit'].scatter(
                dfit2d['lamb'],
                phiflat,
                c=dfit2d['sol_total'][indspect, :],
                s=6, marker='s', edgecolors='None',
                vmin=vmin, vmax=vmax, cmap=cmap,
            )
            if dfit2d['dinput']['symmetry'] is True:
                dax['fit'].axhline(symaxis,
                                   c='k', ls='--', lw=2.)
        if dax.get('fit1d') is not None:
            # dax['fit1d'].plot()
            pass
    else:
        extent = (xi.min(), xi.max(), xj.min(), xj.max())
        if dax.get('err') is not None:
            dax['err'].imshow(err,
                              extent=extent, cmap=plt.cm.seismic,
                              origin='lower', vmin=errm, vmax=errm)
        if dax.get('data') is not None:
            dax['data'].imshow(data,
                               extent=extent, cmap=cmap,
                               origin='lower', vmin=vmin, vmax=vmax)
        if dax.get('fit') is not None:
            dax['fit'].imshow(fit,
                              extent=extent, cmap=cmap,
                              origin='lower', vmin=vmin, vmax=vmax)

    # Plot profiles
    # ------------
    if dax.get('Ti') is not None and dfit2d['dinput']['Ti'] is True:
        for ii in range(dfit2d['kTiev'].shape[1]):
            dax['Ti'].plot(dfit2d['kTiev'][indspect, ii, :]*1.e-3,
                           pts_phi,
                           ls='-', marker='.', ms=4,
                           label=dfit2d['dinput']['width']['keys'][ii])
        dax['Ti'].set_xlim(left=0.)
        dax['Ti'].legend(frameon=True,
                         loc='upper left', bbox_to_anchor=(0., -0.1))

    if dax.get('vi') is not None and dfit2d['dinput']['vi'] is True:
        for ii in range(dfit2d['vims'].shape[1]):
            dax['vi'].plot(dfit2d['vims'][indspect, ii, :]*1.e-3,
                           pts_phi,
                           ls='-', marker='.', ms=4,
                           label=dfit2d['dinput']['shift']['keys'][ii])
        dax['vi'].legend(frameon=True,
                         loc='upper left', bbox_to_anchor=(0., -0.1))

    if dax.get('ratio') not in [None, False]:
        for ii in range(dfit2d['ratio']['value'].shape[1]):
            dax['ratio'].plot(dfit2d['ratio']['value'][indspect, ii, :],
                              pts_phi,
                              ls='-', marker='.', ms=4,
                              label=dfit2d['ratio']['str'][ii])
        dax['ratio'].legend(frameon=True,
                            loc='upper left', bbox_to_anchor=(0., -0.1))

    # Add 1d prof
    # ------------
    if spect1d is not None:
        lcol = ['r', 'k', 'm']
        for ii in range(spect1d.shape[0]):
            l, = dax['fit1d'].plot(lambfit, fit1d[ii, :],
                                   ls='-', lw=1., c=lcol[ii % len(lcol)])
            col = l.get_color()
            dax['fit1d'].plot(lambfit, spect1d[ii, :],
                              ls='None', marker='.', c=col, ms=4)
            dax['err1d'].plot(lambfit, fit1d[ii, :] - spect1d[ii, :],
                              ls='-', lw=1., c=col)
            dax['fit'].axhline(phiminmax[ii][0], c=col, ls='-', lw=2.)
            dax['fit'].axhline(phiminmax[ii][1], c=col, ls='-', lw=2.)

    # double legend
    if dfit2d['dinput']['double'] is not False:
        hand = [mlines.Line2D([], [], c='k', ls='None')]*2
        c0 = (dfit2d['dinput']['double'] is True
              or dfit2d['dinput']['double'].get('dratio') is None)
        if c0:
            dratio = dfit2d['dratio'][indspect]
            dratiostr = ''
        else:
            dratio = dfit2d['dinput']['double']['dratio']
            dratiostr = '  (fixed)'
        c0 = (dfit2d['dinput']['double'] is True
              or dfit2d['dinput']['double'].get('dshift') is None)
        if c0:
            dshift = dfit2d['dshift'][indspect]
            dshiftstr = ''
        else:
            dshift = dfit2d['dinput']['double']['dshift']
            dshiftstr = '  (fixed)'
        lleg = ['dratio = {:4.2f}{}'.format(dratio, dratiostr),
                ('dshift = {:4.2e} * '.format(dshift)
                 + r'$\lambda${}'.format(dshiftstr))]
        legr = dax['err1d'].legend(handles=hand, labels=lleg, title='double',
                                   bbox_to_anchor=(1.01, 0.),
                                   loc='center left')

    # Polishing
    # ------------
    for kax in dax.keys():
        if kax != 'err':
            plt.setp(dax[kax].get_yticklabels(), visible=False)
    if dax.get('fit') is not None:
        plt.setp(dax['fit'].get_xticklabels(), visible=False)
    if dax.get('fit1d') is not None:
        plt.setp(dax['fit1d'].get_xticklabels(), visible=False)
    if dax.get('err_colorbar') is not None and dax['err'] is not None:
        plt.colorbar(errax, ax=dax['err'], cax=dax['err_colorbar'],
                     orientation='horizontal', extend='both')
    if (
        dax.get('data_colorbar') is not None
        and (dax['data'] is not None or dax['fit'] is not None)
    ):
        plt.colorbar(dataax, ax=dax['data'], cax=dax['data_colorbar'],
                     orientation='horizontal')
    return dax


# #################################################################
# #################################################################
#                   noise plot
# #################################################################
# #################################################################

def plot_noise_analysis(dnoise=None, margin=None, fraction=None,
                        ms=None, dcolor=None,
                        dax=None, fs=None, dmargin=None,
                        wintit=None, tit=None, sublab=None,
                        save=None, name=None, path=None, fmt=None):

    # Check inputs
    # ------------
    if save is None:
        save = False

    if fraction is None:
        fraction = 0.4

    if ms is None:
        ms = 4.
    if dcolor is None:
        dcolor = {'mask': (0.4, 0.4, 0.4),
                  'noeval': (0.7, 0.7, 0.7),
                  'S/N': (1., 0., 0.)}
    dcolor['ok'] = dcolor.get('ok', (0., 0., 0.))

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if dmargin is None:
            dmargin = {'left': 0.05, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.95,
                       'wspace': 0.5, 'hspace': None}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
        if sublab is None:
            sublab = True

    # Prepare data
    # ------------
    nbs = dnoise['nbsplines']
    nspect = dnoise['data'].shape[0]
    nlamb = dnoise['chi2n'].shape[1]

    indsort = dnoise['indsort']
    indnan = dnoise['indnan'][:-1]
    phiplot = np.tile(np.insert(dnoise['phi'][indsort[0, :], indsort[1, :]],
                                indnan,
                                np.nan),
                      (nspect, 1)).ravel()
    dataplot = np.insert(dnoise['data'][:, indsort[0, :], indsort[1, :]],
                         indnan,
                         np.nan, axis=1).ravel()
    fitplot = np.insert(dnoise['fit'][:, indsort[0, :], indsort[1, :]],
                        indnan,
                        np.nan, axis=1).ravel()
    errplot = fitplot - dataplot

    # get all indices
    indin = dnoise['indin']
    indout_mask = dnoise['indout_mask']
    indout_noeval = dnoise['indout_noeval']

    # fit sqrt on sigmai
    c0 = (dnoise['var_fraction'] is not False
          or (margin is not None and margin != dnoise['var_margin'])
          or (fraction is not None
              and fraction != dnoise['var_fraction']))
    if c0 is True:
        from . import _fit12d
        (mean, var, xdata, const,
         indout_var, _, margin, _) = _fit12d.get_noise_analysis_var_mask(
             fit=dnoise['fit'], data=dnoise['data'], mask=dnoise['mask'],
             margin=margin, fraction=False)
    else:
        mean = dnoise['var_mean']
        var = dnoise['var']
        const = dnoise['var_const']
        xdata = dnoise['var_xdata']
        indout_var = dnoise['indout_var']
        assert dnoise['var_fraction'] is False

    # Safety check
    indout_var_agg = (np.sum(indout_var, axis=0)/float(indout_var.shape[0])
                      > fraction)
    indout_tot = np.array([indout_mask,
                           indout_noeval,
                           indout_var_agg])
    c0 = np.all(np.sum(indout_tot.astype(int), axis=0) <= 1)
    if not c0:
        msg = "Overlapping indout!"
        raise Exception(msg)

    # get plotting indout
    indout_maskplot = np.tile(np.insert(
        indout_mask[indsort[0, :], indsort[1, :]], indnan, False),
        (nspect, 1)).ravel()
    indout_noevalplot = np.tile(np.insert(
        indout_noeval[indsort[0, :], indsort[1, :]], indnan, False),
        (nspect, 1)).ravel()
    indout_varplot = np.insert(
        indout_var[:, indsort[0, :], indsort[1, :]],
        indnan, False, axis=1).ravel()
    dindout = {
        'mask': {'ind': indout_mask, 'plot': indout_maskplot},
        'noeval': {'ind': indout_noeval},
        # ,'plot': {indout_noevalplot},
        'S/N': {'ind': indout_var_agg, 'plot': indout_varplot},
    }
    indinplot = ~(indout_maskplot | indout_noevalplot | indout_varplot)

    # cam and cmap
    cam = np.zeros(indout_mask.shape, dtype=float)
    for ii, k0 in enumerate(['mask', 'noeval', 'S/N']):
        cam[dindout[k0]['ind']] = ii+1
    cam = cam/4

    cmap = plt.cm.get_cmap('viridis', 4)
    newcolors = cmap(np.linspace(0, 1, 4))
    for ii, k0 in enumerate(['ok', 'mask', 'noeval', 'S/N']):
        newcolors[ii, :] = mcolors.to_rgba(dcolor[k0])
    cmap = ListedColormap(newcolors)

    alpha = 0.2

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(2, 8, **dmargin)
        ax0 = fig.add_subplot(gs[0, 2:4])
        ax1 = fig.add_subplot(gs[0, 4:6], sharey=ax0)
        ax2 = fig.add_subplot(gs[1, 2:4])
        ax3 = fig.add_subplot(gs[1, 4:6])
        ax4 = fig.add_subplot(gs[:, 6:], aspect='equal', adjustable='datalim')
        ax5 = fig.add_subplot(gs[0, 1], sharey=ax0)
        ax6 = fig.add_subplot(gs[1, :2])

        ax0.set_title('Profiles')
        ax0.set_xlabel(r'data (a.u.)')
        ax1.set_title('Err profiles')
        ax1.set_xlabel(r'data (a.u.)')
        ax2.set_title('Error')
        ax2.set_xlabel('fit (a.u.)')
        ax2.set_ylabel('err (a.u.)')
        ax3.set_title('Standard deviation')
        ax3.set_xlabel('data (a.u.)')
        ax3.set_ylabel(r'$\sigma$' + ' (a.u.)')
        ax4.set_title('Camera')
        ax4.set_xlabel(r'$x_i$ (m)')
        ax4.set_ylabel(r'$x_j$ (m)')
        ax5.set_title('bsplines')
        ax5.set_xlabel('')
        ax5.set_ylabel(r'$\phi$ (rad)')
        ax6.set_title('')
        ax6.set_xlabel('mean(fit)')
        ax6.set_ylabel(r'$\chi^2_n$ (a.u.)')

        dax = {'bsplines': ax5,
               'prof': ax0,
               'proferr': ax1,
               'err': ax2,
               'errbin': ax3,
               'cam': ax4,
               'chi2': ax6,
               }

    # Plot main images
    # ------------

    # bsplines
    if dax.get('bsplines') is not None:
        dax['bsplines'].plot(dnoise['bs_val'],
                             np.repeat(dnoise['bs_phi'][:, None],
                                       dnoise['nbsplines'], axis=1),
                             ls='-', lw=1.)

    # Profile
    if dax.get('prof') is not None:
        dax['prof'].plot(dataplot[indinplot], phiplot[indinplot],
                         marker='.', ls='None',
                         mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        dax['prof'].plot(fitplot, phiplot,
                         marker='None', ls='-', lw=1., c='g')
        for k0 in ['mask', 'S/N']:
            dax['prof'].plot(dataplot[dindout[k0]['plot']],
                             phiplot[dindout[k0]['plot']],
                             marker='.', ls='None',
                             mfc=np.r_[dcolor[k0], alpha], mec='None', ms=ms)

    # Profile err
    if dax.get('proferr') is not None:
        dax['proferr'].plot(errplot[indinplot], phiplot[indinplot],
                            marker='.', ls='None',
                            mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        for k0 in ['mask', 'S/N']:
            dax['proferr'].plot(errplot[dindout[k0]['plot']],
                                phiplot[dindout[k0]['plot']],
                                marker='.', ls='None',
                                mfc=np.r_[dcolor[k0], alpha],
                                mec='None', ms=ms)

    # Error distribution
    if dax.get('err') is not None:
        dax['err'].axhline(0., c='k', ls='--', lw=1.)
        lab = r'{}'.format(margin) + r'$\sigma$'
        dax['err'].fill_between(xdata,
                                -margin*const*np.sqrt(xdata),
                                margin*const*np.sqrt(xdata),
                                color=(0.7, 0.7, 0.7, 1.), label=lab)
        dax['err'].plot(fitplot[indinplot], errplot[indinplot],
                        marker='.', ls='None',
                        mfc=(0., 0., 0., alpha), mec='None', ms=ms)
        for k0 in ['mask', 'S/N']:
            dax['err'].plot(fitplot[dindout[k0]['plot']],
                            errplot[dindout[k0]['plot']],
                            marker='.', ls='None',
                            mfc=np.r_[dcolor[k0], alpha], mec='None', ms=ms)
        dax['err'].legend(loc='lower right', frameon=True)

    # Error binning and standard variation estimate
    if dax.get('errbin') is not None:
        dax['errbin'].plot(xdata, np.sqrt(var),
                           marker='.', c='k', ls='None')
        lab = r'$\sigma=$' + r'{:5.3e}'.format(const) + r'$\sqrt{data}$'
        dax['errbin'].plot(xdata, const*np.sqrt(xdata),
                           marker='None', ls='-', c='k', label=lab)
        dax['errbin'].plot(xdata, np.sqrt(xdata),
                           marker='None', ls='--', c='k')
        dax['errbin'].plot(xdata, mean,
                           marker='.', ls='None', c='k')
        dax['errbin'].axhline(0., ls='--', c='k', lw=1., marker='None')
        dax['errbin'].legend(loc='lower right', frameon=True)

    # Camera with identified pixels
    if dax.get('cam') is not None:
        dax['cam'].imshow(cam.T,
                          origin='lower', interpolation='nearest',
                          aspect='equal', cmap=cmap)
        lab = r'err > {:02.0f}% {}$\sigma$'.format(fraction*100, margin)
        hand = [mlines.Line2D([], [], c=dcolor['S/N'], marker='s', ls='None')]
        dax['cam'].legend(hand, [lab], loc='best')

    # chi2
    if dax.get('chi2') is not None:
        dax['chi2'].plot(dnoise['chi2_meandata'].ravel(),
                         (dnoise['chi2n']).ravel(),
                         c='k', marker='.', ms=ms, ls='None')

    # Polish
    # ------------
    if sublab is True:
        if dax.get('bsplines') is not None:
            dax['bsplines'].annotate('(a)',
                                     xy=(0., 1.02),
                                     xycoords='axes fraction',
                                     size=10, fontweight='bold',
                                     horizontalalignment='center',
                                     verticalalignment='bottom')
        if dax.get('prof') is not None:
            dax['prof'].annotate('(b)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('proferr') is not None:
            dax['proferr'].annotate('(c)',
                                    xy=(0., 1.02),
                                    xycoords='axes fraction',
                                    size=10, fontweight='bold',
                                    horizontalalignment='center',
                                    verticalalignment='bottom')
        if dax.get('chi2') is not None:
            dax['chi2'].annotate('(d)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('err') is not None:
            dax['err'].annotate('(e)',
                                xy=(0., 1.02),
                                xycoords='axes fraction',
                                size=10, fontweight='bold',
                                horizontalalignment='center',
                                verticalalignment='bottom')
        if dax.get('errbin') is not None:
            dax['errbin'].annotate('(f)',
                                   xy=(0., 1.02),
                                   xycoords='axes fraction',
                                   size=10, fontweight='bold',
                                   horizontalalignment='center',
                                   verticalalignment='bottom')
        if dax.get('cam') is not None:
            dax['cam'].annotate('(g)',
                                xy=(0., 1.02),
                                xycoords='axes fraction',
                                size=10, fontweight='bold',
                                horizontalalignment='center',
                                verticalalignment='bottom')

    # save
    # ------------
    if save is True:
        if name is None:
            name = 'NoiseAnalysis'
        if fmt is None:
            fmt = 'png'
        if path is None:
            path = './'
        if name[-4:] != '.{}'.format(fmt):
            name = '{}.{}'.format(name, fmt)

        pfe = os.path.join(os.path.abspath(path), name)
        dax['prof'].figure.savefig(pfe, format=fmt)

        msg = "Saved in:\n\t{}".format(pfe)
        print(msg)
    return dax


def plot_noise_analysis_scannbs(
    dnoise=None, ms=None,
    dax=None, fs=None, dmargin=None,
    wintit=None, tit=None, sublab=None,
    save=None, name=None, path=None, fmt=None,
):

    # Check inputs
    # ------------
    if save is None:
        save = False

    if ms is None:
        ms = 2.

    if dax is None:
        if fs is None:
            fs = (18, 9)
        if dmargin is None:
            dmargin = {'left': 0.06, 'right': 0.99,
                       'bottom': 0.06, 'top': 0.96,
                       'wspace': 0.5, 'hspace': 1.}
        if wintit is None:
            wintit = _WINTIT
        if tit is None:
            tit = '2D fitting of X-Ray Crystal Bragg spectrometer'
        if sublab is None:
            sublab = True

    # Prepare data
    # ------------
    nlbs = dnoise['lnbsplines'].size
    nbs = dnoise['nbsplines']
    nspect = dnoise['chi2n'].shape[1]
    nlambu = dnoise['chi2n'].shape[2]

    # Plotting arrays
    lnbsplines_plot = np.repeat(dnoise['lnbsplines'], nspect*nlambu)
    chi2n_plot = dnoise['chi2n'].ravel()
    alpha = np.tile((np.abs(dnoise['dataint'])
                     / np.nanmax(np.abs(dnoise['dataint']))),
                    (nlbs, 1, 1)).ravel()
    colors = np.concatenate((np.tile([0., 0., 0.], (nspect*nlambu*nlbs, 1)),
                             alpha[:, None]),
                            axis=1)
    chi2nmean = (np.nansum(dnoise['chi2n']*dnoise['dataint'][None, :, :],
                           axis=-1).sum(axis=-1)
                 / np.nansum(dnoise['dataint']))
    var_mean = np.sqrt(np.nansum(dnoise['var_mean']**2, axis=1))
    var_std = np.sqrt(np.nansum(
        (np.sqrt(dnoise['var'])
         - dnoise['var_const'][:, None]*np.sqrt(dnoise['var_xdata']))**2,
        axis=1))

    # Prepare figure if dax not provided
    # ------------
    if dax is None:
        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(11, max(1, dnoise['nbsplines'].size*4+2),
                               **dmargin)
        ax0 = fig.add_subplot(gs[:4, :])
        ax0.set_xlabel(r'nb. of bsplines')
        ax0.set_ylabel(r'Residue (a.u.)')
        ax0.set_title('nlamb = {}'.format(dnoise['lambedges'].size-1))

        dax = {'conv': ax0,
               'cases_fit': [],
               'cases_err': [],
               'cases_var': []}
        for ii in range(dnoise['nbsplines'].size):
            if ii == 0:
                dax['cases_fit'].append(fig.add_subplot(
                    gs[5:9, ii*5:ii*5+2]))
                dax['cases_err'].append(fig.add_subplot(
                    gs[5:9, ii*5+2:ii*5+4],
                    sharey=dax['cases_fit'][0]))
                dax['cases_var'].append(fig.add_subplot(
                    gs[9:, ii*5:ii*5+4]))
            else:
                dax['cases_fit'].append(fig.add_subplot(
                    gs[5:9, ii*5:ii*5+2],
                    sharey=dax['cases_fit'][ii-1]))
                dax['cases_err'].append(fig.add_subplot(
                    gs[5:9, ii*5+2:ii*5+4],
                    sharex=dax['cases_err'][0],
                    sharey=dax['cases_fit'][0]))
                dax['cases_var'].append(fig.add_subplot(
                    gs[9:, ii*5:ii*5+4],
                    sharey=dax['cases_var'][ii-1]))
            # dax['cases_err'][ii].
            dax['cases_fit'][ii].set_ylabel(r'$\phi$ (rad)')
            dax['cases_fit'][ii].set_xlabel(r'data (a.u.)')
            dax['cases_err'][ii].set_xlabel(r'error (a.u.)')
            dax['cases_var'][ii].set_xlabel(r'data (a.u.)')
            dax['cases_var'][ii].set_ylabel(r'$\sigma$ (a.u.)')

            dax['cases_fit'][ii].xaxis.set_label_coords(0.5, -0.07)
            dax['cases_err'][ii].xaxis.set_label_coords(0.5, -0.07)
            plt.setp(dax['cases_err'][ii].get_yticklabels(), visible=False)

    # Plot main images
    # ------------

    # Fit chi2n vs nbs
    if dax.get('conv') is not None:
        minus_one = chi2nmean[-1]
        chi2nn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                  * (chi2nmean - minus_one) / (chi2nmean[0] - minus_one)
                  + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], chi2nn,
                         c='k', ls='-', lw=2., label='fit cost')

        # error dictribution moments vs nbs
        varmn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                 * (var_mean - var_mean[-1]) / (var_mean[0] - var_mean[-1])
                 + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], varmn,
                         c='b', ls='-', lw=2., label=r'$\hat{err}_{cost}$')
        varstdn = ((dnoise['var_const'][0] - dnoise['var_const'][-1])
                   * (var_std - var_std[-1]) / (var_std[0] - var_std[-1])
                   + dnoise['var_const'][-1])
        dax['conv'].plot(dnoise['lnbsplines'], varstdn,
                         c='r', ls='-', lw=2., label=r'$\sigma_{cost}$')
        dax['conv'].plot(dnoise['lnbsplines'], dnoise['var_const'],
                         c='g', ls='-', lw=2., label='fitting constant')
        dax['conv'].legend(loc='best')

    # Particular cases
    if dax.get('case_fit') is not None and dnoise.get('nbsplines') is not None:
        inbs = [np.nonzero(dnoise['lnbsplines'] == nbs)[0][0]
                for nbs in dnoise['nbsplines']]
        for ii in range(dnoise['nbsplines'].size):
            indin = dnoise['bs_indin'][ii]
            dax['cases_fit'][ii].plot(dnoise['bs_data'][ii][indin],
                                      dnoise['bs_phidata'][ii][indin],
                                      c=(0.5, 0.5, 0.5),
                                      marker='.', ls='None', ms=ms)
            dax['cases_fit'][ii].plot(dnoise['bs_fit'][ii][indin],
                                      dnoise['bs_phidata'][ii][indin],
                                      c='k', marker='None', ls='-', lw=2.)

            err = (dnoise['bs_fit'][ii] - dnoise['bs_data'][ii])
            dax['cases_err'][ii].scatter(
                err[indin], dnoise['bs_phidata'][ii][indin],
                marker='.', facecolors=(0., 0., 0., 0.5),
                edgecolors='None')
            dax['cases_err'][ii].axvline(0., c='k', ls='--', lw=1.)

            # var
            indok = ~np.isnan(dnoise['var'][inbs[ii], :])
            varfit = (dnoise['var_const'][inbs[ii]]
                      * np.sqrt(dnoise['var_xdata'][indok]))
            lab = (r"{:4.2f}".format(dnoise['var_const'][inbs[ii]])
                   + r"$\sqrt{data}$")
            dax['cases_var'][ii].fill_between(
                dnoise['var_xdata'][indok],
                varfit, np.sqrt(dnoise['var'][inbs[ii], indok]),
                fc=(1., 0., 0., 0.5), ls='-')
            dax['cases_var'][ii].plot(
                dnoise['var_xdata'],
                np.sqrt(dnoise['var'][inbs[ii], :]),
                c='r', marker='.', ls='None', ms=ms)
            dax['cases_var'][ii].plot(
                dnoise['var_xdata'][indok], varfit,
                c='g', marker='None', ls='-', label=lab)

            # mean
            dax['cases_var'][ii].fill_between(
                dnoise['var_xdata'], 0.,
                dnoise['var_mean'][inbs[ii], :],
                fc=(0., 0., 1., 0.5), ls='-')
            dax['cases_var'][ii].axhline(0., c='k', ls='--', lw=1.)
            dax['cases_var'][ii].plot(dnoise['var_xdata'],
                                      dnoise['var_mean'][inbs[ii], :],
                                      c='b', ls='-', lw=1.)
            dax['cases_var'][ii].legend(loc='upper left')

            posx = 1.1
            posy = 1.05
            plt.text(posx, posy, '{} bsplines'.format(dnoise['nbsplines'][ii]),
                     horizontalalignment='center',
                     verticalalignment='center',
                     size=12, fontweight='bold',
                     transform=dax['cases_fit'][ii].transAxes)

            if dax.get('conv') is not None:
                dax['conv'].axvline(dnoise['nbsplines'][ii],
                                    c='k', ls='--', lw=1.)

    # Polish
    # ------------
    if sublab is True:
        if dax.get('conv') is not None:
            dax['conv'].annotate('(a)',
                                 xy=(0., 1.02),
                                 xycoords='axes fraction',
                                 size=10, fontweight='bold',
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
        if dax.get('cases_fit') is not None:
            ls = ['b', 'c', 'd', 'e', 'f']
            for ii in range(len(dax['cases_fit'])):
                dax['cases_fit'][ii].annotate(
                    '({})'.format(ls[ii]),
                    xy=(0., 1.02), xycoords='axes fraction',
                    size=10, fontweight='bold',
                    horizontalalignment='center',
                    verticalalignment='bottom')

    # save
    # ------------
    if save is True:
        if name is None:
            name = 'NoiseAnalysis'
        if fmt is None:
            fmt = 'png'
        if path is None:
            path = './'
        if name[-4:] != '.{}'.format(fmt):
            name = '{}.{}'.format(name, fmt)

        pfe = os.path.join(os.path.abspath(path), name)
        dax['conv'].figure.savefig(pfe, format=fmt)

        msg = "Saved in:\n\t{}".format(pfe)
        print(msg)
    return dax